#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk screen plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; A Gtk-based <screen> for Edwin.
;;; Package: (edwin screen gtk-screen)

(define-class (<gtk-screen>
	       (constructor %make-gtk-screen
			    (toplevel editor-thread event-queue) no-init))
    (<screen>) ;; TODO: could also be a <gtk-window>, replacing toplevel!

  ;; The toplevel <gtk-window>.  The top widget.
  (toplevel define accessor)

  ;; The thread to interrupt when ^G is seen by an event handler (i.e. a
  ;; callback in the glib-thread).
  (editor-thread define accessor)

  ;; The editor-thread's event queue.
  (event-queue define accessor)

  ;; An alist of Edwin buffers and their drawings, to be shared among
  ;; the text-widgets, and updated during screen update.
  (drawings define standard initial-value '())

  ;; The window/icon/taskbar name.  #f just means "not set".  Cannot be
  ;; set to #f!
  (name define standard initial-value #f)

  ;; Toplevel's gtk-style-context, after toplevel is realized.
  (style-context define standard initial-value #f)

  ;; The default font (description) and its character dimensions.
  (font define standard)
  (char-width define standard)
  (line-height define standard)
  (line-spacing define standard)

  ;; The thread that blinks the cursor and the blinking <cursor-ink>.
  (blinker define standard)
  (blinking define standard initial-value #f)

  ;; Whether a cursor should be blinking.
  (in-focus? define standard initial-value #f))

(define screen-list)			;serialize with glib lock

(define (make-gtk-screen #!optional geometry)
  (%trace "; make-gtk-screen "geometry"\n")
  (with-glib-lock
   (lambda ()
     (let* ((toplevel (gtk-window-new 'toplevel))
	    (screen (%make-gtk-screen toplevel (current-thread) (make-queue)))
	    (geometry* (if (default-object? geometry)
			   "80x24"
			   (begin
			     (guarantee-string geometry 'make-gtk-screen)
			     geometry))))
       (gtk-widget-set-opacity toplevel 0.95)
       (init-font-dimensions! screen "Monospace 11")
       (init-size! screen geometry*)
       (let ((thread (create-blinker-thread screen)))
	 (%trace ";   blinker thread: "thread"\n")
	 (set-gtk-screen-blinker! screen thread)
	 (detach-thread thread)
	 (%trace ";   editor thread: "(current-thread)"\n"))
       (set! screen-list (cons screen screen-list))
       (%trace ";   screen: "screen"\n")
       screen))))

(define (init-font-dimensions! screen spec)
  (%trace ";   init-font-dimensions! "screen" "spec"\n")
  ;; Lookup screen's font via the toplevel widget's pango-context,
  ;; which appears to be available before toplevel is realized.

  (let* ((toplevel (gtk-screen-toplevel screen))
	 (context (gtk-widget-get-pango-context toplevel))
	 (font (pango-font-description-from-string spec))
	 (metrics (pango-context-get-metrics context font)))
    (set-gtk-screen-font! screen font)
    (let ((ascent (pangos->pixels (pango-font-metrics-get-ascent metrics)))
	  (descent (pangos->pixels (pango-font-metrics-get-descent metrics)))
	  (width (pangos->pixels
		  (pango-font-metrics-get-approximate-char-width metrics)))
	  (spacing (pangos->pixels (pango-context-spacing context))))
      (if (zero? width)
	  (error "could not get metrics for font" spec))
      (set-gtk-screen-char-width! screen width)
      (set-gtk-screen-line-spacing! screen spacing)
      (set-gtk-screen-line-height! screen (fix:+ ascent descent))
      (%trace ";     Font: \""spec"\" "width"x"ascent"+"descent" "spacing"\n")
      (pango-font-metrics-unref metrics))))

(define (init-size! screen geometry)
  (%trace ";   init-size! "screen" "geometry"\n")
  ;; Sets the logical screen size.  This sets Edwin window and thus
  ;; text-widget sizes, which ultimately determine the GtkWindow size
  ;; request.
  (parse-geometry
   geometry
   (lambda (width height x y)
     (declare (ignore x y))
     ;; For make-editor-frame:
     (set-screen-x-size! screen width)
     (set-screen-y-size! screen height))))

(define (parse-geometry geometry receiver)
  (let* ((num "[0-9]+")
	 (size-patt (string "\\("num"\\)x\\("num"\\)"))
	 (position-patt (string "\\([-+]"num"\\)\\([-+]"num"\\)"))
	 (extract (lambda (regs index)
		    (string->number (re-match-extract geometry regs index)))))
    (declare (integrate extract))
    (cond ((re-string-match (string size-patt position-patt) geometry)
	   => (lambda (regs)
		(receiver (extract regs 1) (extract regs 2)
			  (extract regs 3) (extract regs 4))))
	  ((re-string-match position-patt geometry)
	   => (lambda (regs)
		(receiver #f #f
			  (extract regs 1) (extract regs 2))))
	  ((re-string-match size-patt geometry)
	   => (lambda (regs)
		(receiver (extract regs 1) (extract regs 2)
			  #f #f)))
	  (else
	   (error:wrong-type-argument geometry
				      "window geometry (e.g. \"80x40-0-0\")"
				      'parse-geometry)))))

(define (x-size->width screen x-size)
  (fix:* x-size (gtk-screen-char-width screen)))

(define (y-size->height screen y-size)
  (fix:+ (fix:* y-size (gtk-screen-line-height screen))
	 (fix:* (fix:1+ y-size) (gtk-screen-line-spacing screen))))

(define (width->x-size screen width)
  (fix:quotient width (gtk-screen-char-width screen)))

(define (height->y-size screen height)
  (let ((line-spacing (gtk-screen-line-spacing screen))
	(line-height (gtk-screen-line-height screen)))
    (fix:quotient (fix:- height line-spacing)
		  (fix:+ line-height line-spacing))))

(define (column->x screen column)
  (fix:* column (gtk-screen-char-width screen)))

(define (row->y screen row)
  (fix:* row (fix:+ (gtk-screen-line-spacing screen)
		    (gtk-screen-line-height screen))))

(define (window-text-widget* window)
  (any-child (lambda (widget)
	       (and (text-widget? widget)
		    (eq? window (text-widget-buffer-frame widget))))
	     (gtk-screen-toplevel (window-screen window))))

(define-integrable (window-modeline* window)
  (let ((widget (window-text-widget* window)))
    (and widget (text-widget-modeline widget))))

(define-integrable (window-cursor-ink* window)
  (let ((widget (window-text-widget* window)))
    (and widget (text-widget-cursor-ink widget))))

(define-integrable (selected-text-widget* screen)
  (let ((window (screen-cursor-window screen)))
    (and window (window-text-widget* window))))

(declare (integrate-operator car*))
(define (car* obj) (and (pair? obj) (car obj)))

(declare (integrate-operator cdr*))
(define (cdr* obj) (and (pair? obj) (cdr obj)))

(define-method set-screen-size! ((screen <gtk-screen>) x-size y-size)
  (%trace "; (set-screen-size! <gtk-screen>) "screen" "x-size"x"y-size"\n")
  (without-interruption
   (lambda ()
     (set-screen-x-size! screen x-size)
     (set-screen-y-size! screen y-size)
     (send (screen-root-window screen) ':set-size! x-size y-size))))

(define %trace-blinker? #f)

(define (create-blinker-thread screen)

  (define-syntax %trace3
    (syntax-rules ()
      ((_ args ...) (if %trace-blinker? (outf-error args ...)))))

  (create-thread
   #f
   (lambda ()
     (%trace2 ";blinking started on "screen"\n")
     (let loop ()
       (without-interruption
	(lambda ()
	  (let ((cursor (gtk-screen-blinking screen)))
	    (cond ((not cursor)
		   (%trace2 ";blinker: no blinking "screen"\n")
		   (suspend-current-thread)
		   (%trace2 ";blinker: awake after not blinking "screen"\n"))
		  ((not (cursor-ink-visible? cursor))
		   (%trace2 ";blinker: invisible "cursor"\n")
		   (suspend-current-thread)
		   (%trace2 ";blinker: awake after invisible "cursor"\n"))
		  (else
		   (%trace3 ";blinker: off "cursor"\n")
		   (with-glib-lock
		    (lambda () (set-fix-ink-widgets! cursor '())))
		   (sleep-current-thread 500)
		   (if (cursor-ink-visible? cursor)
		       (begin
			 (%trace3 ";blinker: on "cursor"\n")
			 (with-glib-lock
			  (lambda ()
			    (set-fix-ink-widgets!
			     cursor (cursor-ink-widget-list cursor))))
			 (sleep-current-thread 500))
		       (begin
			 (%trace ";blinker: on: invisible "cursor"\n")
			 unspecific)))))))
       (loop)))))

(define-method screen-beep ((screen <gtk-screen>))
  (assert-glib-locked '(screen-beep <gtk-screen>))
  (gtk-widget-error-bell (gtk-screen-toplevel screen)))

(define-method screen-enter! ((screen <gtk-screen>))
  (%trace "; screen-enter! "screen"\n")
  (assert-glib-locked '(screen-enter! <gtk-screen>))
  (update-widgets screen)
  (gtk-window-present (gtk-screen-toplevel screen))
  (%trace "; screen-enter!: done\n"))

(define-method screen-exit! ((screen <gtk-screen>))
  (%trace "; screen-exit! "screen"\n")
  (assert-glib-locked '(screen-exit! <gtk-screen>))
  (set-gtk-screen-in-focus?! screen #f)
  (update-blinking screen))

(define-method screen-discard! ((screen <gtk-screen>))
  (%trace "; screen-discard! "screen"\n")
  (assert-glib-locked '(screen-discard! <gtk-screen>))
  (set! screen-list (delq! screen screen-list))
  (gtk-widget-destroy (gtk-screen-toplevel screen))
  (pango-font-description-free (gtk-screen-font screen)))

(define-method screen-modeline-event! ((screen <gtk-screen>) window type)
  (%trace "; screen-modeline-event! "screen" "window" "type"\n")
  (assert-glib-locked '(screen-modeline-event! <gtk-screen>))
  unspecific)

;;; These scrolling procedures are for editor commands (not
;;; scrollbars).  They force a buffer-drawing layout update
;;; (effectively, a redisplay) after which they can map window coords
;;; to drawing coords to line ink to buffer index.

(define-method screen/window-scroll-y-absolute! ((screen <gtk-screen>)
						 frame y-point)
  (%trace "; screen/window-scroll-y-absolute! "screen" "frame" "y-point"\n")
  (assert-glib-locked '(screen/window-scroll-y-absolute! <gtk-screen>))
  (with-updated-window
   screen frame 'scroll-y-absolute!
   (lambda (widget)
     (let ((cursor (text-widget-cursor-ink widget))
	   (view (fix-layout-view widget)))
       (let ((desired-y (fix:+ (fix-rect-y view)
			       (row->y screen y-point)))
	     (actual-y (fix-rect-y (fix-ink-extent cursor))))
	 (%scroll-to screen widget
		     (fix-rect-x view)
		     (fix:+ (fix:- actual-y desired-y)
			    (fix-rect-y view))))))))

(define (%scroll-to screen widget x y)
  (let* ((max-y (let ((drawing (text-widget-buffer-drawing widget)))
		  (if drawing
		      (fix:max 0
			       (fix:- (fix-rect-max-y
				       (fix-drawing-extent drawing))
				      (gtk-screen-line-height screen)))
		      0)))
	 (y* (fix:min max-y (fix:max 0 y))))
    (%trace ";   %scroll-to "x" "y*"\n")
    (fix-layout-scroll-to! widget x y*)
    (update-start-mark widget)))

(define (with-updated-window screen frame what operation)
  (%trace "; with-updated-window "screen" "frame" "what"\n")

  (if (not (screen-in-update? screen))
      ;; Don't loop when used during screen update(!).
      (begin
	(%trace ";   forcing update...\n")
	(update-screens! #t)
	(%trace ";   ...forced update finished.\n"))
      (%trace ";   in update, with widget "(window-text-widget* frame)"\n"))

  (let ((widget (window-text-widget* frame)))
    (if (not widget) (error "No widget:" frame))
    (operation widget)))

(define-method screen/window-scroll-y-relative! ((screen <gtk-screen>)
						 frame delta)
  (%trace "; screen/window-scroll-y-relative! "screen" "frame" "delta"\n")
  (assert-glib-locked '(screen/window-scroll-y-relative! <gtk-screen>))
  (with-updated-window
   screen frame 'scroll-y-relative!
   (lambda (widget)
     (let ((view (fix-layout-view widget))
	   (delta* (row->y screen delta)))
       (%scroll-to screen widget
		   (fix-rect-x view)
		   (fix:+ delta* (fix-rect-y view)))
       (update-point widget)))))

(define-method screen/set-window-start-mark! ((screen <gtk-screen>)
					      frame mark force?)
  (%trace "; screen/set-window-start-mark! "screen" "frame" "mark" "force?"\n")
  (assert-glib-locked '(screen/set-window-start-mark! <gtk-screen>))
  (with-updated-window
   screen frame 'set-start-mark!
   (lambda (widget)
     (let ((view (fix-layout-view widget))
	   (line (find-line-at mark widget)))
       (let ((x (fix-rect-x view))
	     (y (if line
		    (fix:- (fix-rect-y (fix-ink-extent line))
			   (gtk-screen-line-spacing screen))
		    0)))
	 (cond (force?
		(fix-layout-scroll-to! widget x y)
		(update-start-mark widget)
		(update-point widget))
	       ((let ((extent (fix-ink-extent (text-widget-cursor-ink widget))))
		  (and (fix:<= y
			       (fix-rect-min-y extent))
		       (fix:< (fix-rect-min-y extent)
			      (fix:+ y (fix-rect-height view)))))
		(fix-layout-scroll-to! widget x y)
		(update-start-mark widget))))))))

(define-method screen/window-mark-visible? ((screen <gtk-screen>) frame mark)
  (%trace "; screen/window-mark-visible? "screen" "frame" "mark"\n")
  (assert-glib-locked '(screen/window-mark-visible? <gtk-screen>))
  (with-updated-window
   screen frame 'mark-visible?
   (lambda (widget)
     (let ((view (fix-layout-view widget))
	   (line (find-line-at mark widget)))
       (let ((min-y (if line
			(fix-rect-min-y (fix-ink-extent line))
			0)))
	 (if (and (fix:<= (fix-rect-min-y view)
			  min-y)
		  (fix:< min-y
			 (fix-rect-max-y view)))
	     (begin
	       (%trace ";   visible\n")
	       #t)
	     (begin
	       (%trace ";   NOT visible\n")
	       #f)))))))

(define-method screen/window-mark->x ((screen <gtk-screen>) frame mark)
  (%trace "; screen/window-mark->x "screen" "frame" "mark"\n")
  (assert-glib-locked '(screen/window-mark->x <gtk-screen>))
  (with-updated-window
   screen frame 'mark->x
   (lambda (widget)
     (let ((line (find-line-at mark widget)))
       (if (not line)
	   0
	   (line->column mark line widget screen))))))

(define-integrable (line->column mark line widget screen)
  (let ((view (fix-layout-view widget))
	(column-width (gtk-screen-char-width screen)))
    (fix:- (image-column mark line)
	   (fix:quotient (fix-rect-x view) column-width))))

(define-method screen/window-mark->y ((screen <gtk-screen>) frame mark)
  (%trace "; screen/window-mark->y "screen" "frame" "mark"\n")
  (assert-glib-locked '(screen/window-mark->y <gtk-screen>))
  (with-updated-window
   screen frame 'mark->y
   (lambda (widget)
     (let ((line (find-line-at mark widget)))
       (if (not line)
	   0
	   (line->row line widget screen))))))

(declare (integrate-operator line->row))
(define (line->row line widget screen)
  (let ((view (fix-layout-view widget))
	(row-height (fix:+ (gtk-screen-line-spacing screen)
			   (gtk-screen-line-height screen))))
    (fix:quotient (fix:- (fix-rect-y (fix-ink-extent line))
			 (fix-rect-y view))
		  row-height)))

(define-method screen/window-mark->coordinates ((screen <gtk-screen>)
						frame mark)
  (%trace "; screen/window-mark->coordinates "screen" "frame" "mark"\n")
  (assert-glib-locked '(screen/window-mark->coordinates <gtk-screen>))
  (with-updated-window
   screen frame 'mark->coordinates
   (lambda (widget)
     (let ((line (find-line-at mark widget)))
       (if (not line)
	   (cons 0 0)
	   (cons (line->column mark line widget screen)
		 (line->row line widget screen)))))))

(define-method screen/window-point-x ((screen <gtk-screen>) frame)
  (assert-glib-locked '(screen/window-point-x <gtk-screen>))
  (screen/window-mark->x screen frame (window-point frame)))

(define-method screen/window-point-y ((screen <gtk-screen>) frame)
  (assert-glib-locked '(screen/window-point-y <gtk-screen>))
  (screen/window-mark->y screen frame (window-point frame)))

(define-method screen/window-point-coordinates ((screen <gtk-screen>) frame)
  (assert-glib-locked '(screen/window-point-coordinates <gtk-screen>))
  (screen/window-mark->coordinates screen frame (window-point frame)))

(define-method screen/window-coordinates->mark ((screen <gtk-screen>)
						frame x y)
  (%trace "; screen/window-coordinates->mark "screen" "frame" "x" "y"\n")
  (%trace-buttons "coordinates->mark "screen" "frame" "x" "y)
  (assert-glib-locked '(screen/window-coordinates->mark! <gtk-screen>))
  (with-updated-window
   screen frame 'coordinates->mark
   (lambda (widget)
     (let ((drawing (fix-layout-drawing widget))
	   (view (fix-layout-view widget))
	   (column-width (gtk-screen-char-width screen))
	   (row-height (fix:+ (gtk-screen-line-spacing screen)
			      (gtk-screen-line-height screen))))
       (let ((buffer (buffer-drawing-buffer drawing))
	     (drawing-x (fix:+ (fix:* x column-width) (fix-rect-x view)))
	     (drawing-y (fix:+ (fix:* y row-height) (fix-rect-y view))))
	 (%trace-buttons "drawing y "drawing-y)
	 (let ((line (find-line-after drawing-y widget)))
	   (%trace-buttons "line after "drawing-y": "line)
	   (cond ((not line)
		  (%trace-buttons "empty buffer")
		  (mark-temporary-copy (buffer-start buffer)))
		 ((fix:< (fix-rect-max-y (fix-ink-extent line))
			 drawing-y)	;mouse was beyond the last line
		  (%trace-buttons "past last line")
		  (mark-temporary-copy (buffer-end buffer)))
		 (else
		  (let ((index
			 (vector-ref
			  (group-column->index
			   (buffer-group buffer)
			   (line-ink-start-index line)
			   (line-ink-end-index line)
			   0 ;; start column
			   (fix:quotient drawing-x column-width)
			   (buffer-drawing-tab-width drawing)
			   (buffer-drawing-char-image-strings drawing))
			  0)))
		    (%trace-buttons "index at "(fix:quotient drawing-x column-width)": "index)
		    (make-mark (buffer-group buffer) index))))))))))

(declare (integrate-operator update-start-mark))
(define (update-start-mark widget)
  ;; Set widget's window's start-mark to the start of the first
  ;; completely visible line ink.
  (let ((line (find-line-after (fix-rect-y (fix-layout-view widget)) widget)))
    (move-mark-to! (get-start-mark widget)
		   (line-start line widget))))

(declare (integrate-operator get-start-mark))
(define (get-start-mark widget)
  (let ((window (frame-text-inferior (text-widget-buffer-frame widget))))
    (or (%window-start-mark window)
	(let ((new (mark-permanent-copy (no-line-start widget))))
	  (%set-window-start-mark! window new)
	  new))))

(define-integrable (no-line-start widget)
  (buffer-drawing-display-start (fix-layout-drawing widget)))

(define (update-point widget)
  (%trace "; update-point "widget"\n")
  ;; Move widget's window's point into view at the beginning of the
  ;; nearest (first or last) completely visible line.

  (declare (integrate-operator move-point))
  (define (move-point for/back line)
    (let ((window (frame-text-inferior (text-widget-buffer-frame widget))))
      (%trace ";   "for/back"ward to "line"\n")
      (%set-window-point-index! window
				(mark-index (line-start line widget)))
      (%set-window-point-moved?! window #t)
      (update-cursor widget)))

  (let ((extent (fix-ink-extent (text-widget-cursor-ink widget)))
	(view (fix-layout-view widget)))
    (cond ((fix:< (fix-rect-min-y extent)
		  (fix-rect-min-y view))
	   (move-point 'for (find-line-after (fix-rect-min-y view) widget)))
	  ((fix:< (fix-rect-max-y view)
		  (fix-rect-max-y extent))
	   (move-point 'back (find-line-before (fix-rect-max-y view) widget)))
	  (else
	   (%trace ";   no need to move\n")))))

(declare (integrate-operator line-start))
(define (line-start line widget)
  (if line
      (line-ink-start line)
      (no-line-start widget)))

(define (find-line-at point widget)
  ;; Return the line-ink that includes the character at point.  If
  ;; there is no such line, return #f or the last line found.
  (let loop ((inks (fix-drawing-display-list
		     (fix-layout-drawing widget)))
	     (last #f))
    (cond ((null? inks) last)
	  ((not (line-ink? (car inks)))
	   (loop (cdr inks) last))
	  (else
	   (let ((line (car inks)))
	     (if (mark< point (line-ink-end line))
		 line
		 (loop (cdr inks) line)))))))

(define (find-line-after y widget)
  ;; Find the first line-ink that starts at or below Y, or the last
  ;; (closest) line.  Returns #f when the buffer is empty.
  (let loop ((inks (fix-drawing-display-list (fix-layout-drawing widget)))
	     (previous #f))
    (if (pair? inks)
	(let ((ink (car inks)))
	  (if (line-ink? ink)
	      (if (fix:<= y (fix-rect-y (fix-ink-extent ink)))
		  ink
		  (loop (cdr inks) ink))
	      (loop (cdr inks) previous)))
	previous)))

(define (find-line-before y widget)
  ;; Find the last line-ink that ends at or above Y.  Returns #f when
  ;; the buffer is empty.
  (let loop ((inks (fix-drawing-display-list (fix-layout-drawing widget)))
	     (previous #f))
    (if (pair? inks)
	(let ((ink (car inks)))
	  (if (line-ink? ink)
	      (if (fix:< y (fix-rect-max-y (fix-ink-extent ink)))
		  previous
		  (loop (cdr inks) ink))
	      (loop (cdr inks) previous)))
	previous)))

;;; Event Handling

(define (queue-input-event screen input-event)
  (%trace2 ";queue-input-event "screen" "input-event"\n")
  (let ((queue (gtk-screen-event-queue screen)))
    (signal-thread-event (gtk-screen-editor-thread screen)
      (named-lambda (gtk-screen-enqueue)
	(%trace2 ";queue-input event "screen" "input-event"\n")
	(enqueue!/unsafe queue input-event)))))

(define (block-for-event-until event-queue time)
  (%trace2 ";block-for-event-until\n")
  (let ((output-available? #f)
	(timeout? #f)
	(registrations))
    (if (and time (not (zero? time)))
	(begin
	  (%trace2 ";block-for-event-until setting timer\n")
	  (register-timer-event (- time (real-time-clock))
				(lambda ()
				  (%trace2
				   ";block-for-event-until timer expired\n")
				  (set! timeout? #t)))))
    (dynamic-wind
     (lambda ()
       (%trace2 ";block-for-event-until registering IO events\n")
       (let ((thread (current-thread)))
	 (set! registrations
	       (register-process-output-events
		thread (lambda (mode)
			 mode
			 (set! output-available? #t))))))
     (lambda ()
       (let loop ()
	 (%trace2 ";block-for-event-until blocking\n")
	 (with-thread-events-blocked
	  (lambda ()
	    (if (and (queue-empty? event-queue)
		     (not (process-status-changes?))
		     (not inferior-thread-changes?)
		     (not output-available?)
		     (not timeout?))
		(suspend-current-thread))))
	 (cond ((not (queue-empty? event-queue))
		(%trace2 ";block-for-event-until input-event\n")
		'input-event)
	       (output-available?
		(%trace2 ";block-for-event-until process-output\n")
		'process-output)
	       (inferior-thread-changes?
		(%trace2 ";block-for-event-until inferior-thread-output\n")
		'inferior-thread-output)
	       ((process-status-changes?)
		(%trace2 ";block-for-event-until process-status\n")
		'process-status)
	       (timeout?
		'timeout)
	       (else
		(loop)))))
     (lambda ()
       (%trace2 ";block-for-event-until deregistering\n")
       (for-each deregister-io-thread-event registrations)
       (set! registrations)))))

(define (probe-for-event event-queue)
  (%trace2 ";probe-for-event\n")
  (cond ((not (queue-empty? event-queue))
	 (%trace2 ";probe-for-event input-event\n")
	 'input-event)
	((process-output-available?)
	 (%trace2 ";probe-for-event process-output\n")
	 'process-output)
	(inferior-thread-changes?
	 (%trace2 ";probe-for-event inferior-thread-output\n")
	 'inferior-thread-output)
	((process-status-changes?)
	 (%trace2 ";probe-for-event process-status\n")
	 'process-status)
	(else
	 (%trace2 ";probe-for-event none\n")
	 'timeout)))

(define (block-for-input-event event-queue msec)
  (let ((time (and msec (not (zero? msec))
		   (queue-empty? event-queue)
		   (+ (real-time-clock)
		      (internal-time/seconds->ticks
		       (/ msec 1000))))))
    (let loop ()
      (or (not (queue-empty? event-queue))
	  (case (if (and msec (zero? msec))
		    (probe-for-event event-queue)
		    (block-for-event-until event-queue time))
	    ((input-event)
	     #t)
	    ((process-status)
	     (with-glib-lock
	      (lambda ()
		(if (handle-process-status-changes)
		    (update-screens! #f))))
	     (loop))
	    ((process-output)
	     (with-glib-lock
	      (lambda ()
		(if (accept-process-output)
		    (update-screens! #f))))
	     (loop))
	    ((inferior-thread-output)
	     (with-glib-lock
	      (lambda ()
		(if (accept-thread-output)
		    (update-screens! #f))))
	     (loop))
	    ((timeout)
	     #f)
	    (else (error "Unexpected value.")))))))

(define-integrable (queue/peek queue)
  (let ((objects (cadr queue)))
    (and (pair? objects)
	 (car objects))))

(declare (integrate-operator queue/push!))
(define (queue/push! queue object)
  (let ((next (cons object (cadr queue))))
    (set-car! (cdr queue) next)
    (if (not (pair? (cddr queue)))
	(set-cdr! (cdr queue) next))))

(define (get-gtk-input-operations screen)
  (let ((event-queue (gtk-screen-event-queue screen)))
    (values
     (named-lambda (gtk-screen-halt-update?)
       ;; Large buffers will generate large runs of these traces...
       ;;(%trace2 ";halt-update?")
       (let ((halt? (not (queue-empty? event-queue))))
	 ;;(%trace2 " => "halt?"\n")
	 halt?))
     (named-lambda (gtk-screen-peek-no-hang msec)
       (%trace2 ";peek-no-hang "msec"\n")
       (without-glib-lock
	(lambda ()
	  (let ((event (and (block-for-input-event event-queue msec)
			    (queue/peek event-queue))))
	    (%trace2 ";peek-no-hang "msec" => "event"\n")
	    event))))
     (named-lambda (gtk-screen-peek)
       (%trace2 ";peek\n")
       (without-glib-lock
	(lambda ()
	  (let ((event (and (block-for-input-event event-queue #f)
			    (queue/peek event-queue))))
	    (%trace2 ";peek => "event"\n")
	    event))))
     (named-lambda (gtk-screen-read)
       (%trace2 ";read\n")
       (without-glib-lock
	(lambda ()
	  (let ((event (and (block-for-input-event event-queue #f)
			    (dequeue!/unsafe event-queue))))
	    (%trace2 ";read => "event"\n")
	    event)))))))

(set!
 os/interprogram-cut
 (named-lambda (os/interprogram-cut string point)
   (declare (ignore point))
   (assert-glib-locked 'os/interprogram-cut)
   (gtk-window-set-clipboard-text (gtk-screen-toplevel (selected-screen))
				  string)))

(set!
 os/interprogram-paste
 (named-lambda (os/interprogram-paste point)
   (declare (ignore point))
   (assert-glib-locked 'os/interprogram-paste)
   (gtk-window-get-clipboard-text (gtk-screen-toplevel (selected-screen)))))

(define (with-gtk-grabbed receiver)
  (%trace "; with-gtk-grabbed "(current-thread)"\n")
  (with-glib-lock (lambda () (receiver with-gtk-ungrabbed '()))))

(define (with-gtk-ungrabbed thunk)
  (%trace "; with-gtk-ungrabbbed "(current-thread)"\n")
  (without-glib-lock thunk))

(define (focus-change-handler widget in?)
  (%trace "; Focus-"(if in? "in" "out")": "widget"\n")
  (let ((screen (edwin-widget-screen widget)))
    (set-gtk-screen-in-focus?! screen in?)
    (update-blinking screen))
  #f)

(define (visibility-notify-handler widget state)
  (%trace "; Visibility: "state" "widget"\n")
  (let ((screen (edwin-widget-screen widget)))
    (case state
      ((visible) (set-screen-visibility! screen 'visible))
      ((partially-obscured) (set-screen-visibility! screen 'partially-obscured))
      ((obscured) (set-screen-visibility! screen 'obscured))
      (else (warn "unexpected visibility state:" state))))
  #t)

(define (key-press-handler widget key char-bits)
  (%trace "; Key-press: "key" "char-bits" "widget"\n")
  (let* ((screen (edwin-widget-screen widget))
	 (queue! (lambda (x)
		   (queue-input-event screen x)
		   (%trace ";  queued "x"\n")
		   #t))
	 (k (case key
	      ((#\backspace) #\rubout)
	      ((#\rubout) #\c-d)
	      ((#\return) #\c-m)
	      ((#\linefeed) #\c-j)
	      ((#\tab) #\c-i)
	      ((|Shift_L| |Shift_R| |Control_L| |Control_R|
		|Caps_Lock| |Shift_Lock|
		|Meta_L| |Meta_R| |Alt_L| |Alt_R|
		|Super_L| |Super_R| |Hyper_L| |Hyper_R|)
	       #f)
	      (else (if (symbol? key)
			(intern (symbol->string key))
			key)))))
    (if (char? k)
	(if (char=? k #\BEL)
	    (let* ((screen (edwin-widget-screen widget))
		   (thread (gtk-screen-editor-thread screen)))
	      (%trace ";  pushing ^G in "(current-thread)"...\n")
	      (queue/push! (gtk-screen-event-queue screen) #\alarm)
	      (%trace ";  signaling "thread"\n")
	      (signal-thread-event
		  thread
		(lambda ()
		  (%trace ";interrupt! in editor "(current-thread)"\n")
		  (editor-beep)
		  (temporary-message "Quit")
		  (^G-signal)))
	      (%trace ";  pushed ^G in "(current-thread)".\n")
	      #t)
	    (queue! (merge-bucky-bits k char-bits)))
	(if k
	    (queue! (make-special-key k char-bits))
	    #t))))

(define (button-down-handler widget type button modifiers x y)
  (declare (ignore type))
  (%trace-buttons "down "button" "modifiers" "x" "y" "widget)
  (queue-input-event (edwin-widget-screen widget)
		     (make-input-event
		      'button
		      execute-gtk-button-command
		      widget button modifiers x y))
  #t)

(define (execute-gtk-button-command widget button modifiers x y)
  (%trace-buttons "execute "widget" "button" "modifiers" "x" "y)
  (let ((screen (edwin-widget-screen widget))
	(frame (text-widget-buffer-frame widget)))
    (let ((inferior (find-inferior (window-inferiors (window-superior frame))
				   frame)))
      (%trace-buttons "window position"
		      " " (fix:quotient x (gtk-screen-char-width screen))
		      "," (fix:quotient
			   y
			   (fix:+ (gtk-screen-line-height screen)
				  (gtk-screen-line-spacing screen))))
      (%trace-buttons "inferior "inferior)
      (%trace-buttons "screen position"
		      " " (fix:+ (inferior-x-start inferior)
				 (fix:quotient x (gtk-screen-char-width screen)))
		      "," (fix:+ (inferior-y-start inferior)
				 (fix:quotient y
					       (fix:+ (gtk-screen-line-height screen)
						      (gtk-screen-line-spacing screen)))))
      (execute-button-command
       screen
       (make-down-button (fix:-1+ button) (modifiers->char-bits modifiers))
       (fix:+ (inferior-x-start inferior)
	      (fix:quotient x
			    (gtk-screen-char-width screen)))
       (fix:+ (inferior-y-start inferior)
	      (fix:quotient y
			    (fix:+ (gtk-screen-line-height screen)
				   (gtk-screen-line-spacing screen))))))))

(define (modifiers->char-bits modifiers)
  (reduce bitwise-ior 0 (map (lambda (modifier)
			       (case modifier
				 ((meta) char-bit:meta)
				 ((control) char-bit:control)
				 ((super) char-bit:super)
				 ((hyper) char-bit:hyper)
				 (else 0)))
			     modifiers)))

;;; Initialization

(define gtk-display-type)

(define (initialize-package!)
  (set! screen-list '())
  (set! gtk-display-type
	(make-display-type 'gtk
			   #t
			   gtk-initialized?
			   make-gtk-screen
			   get-gtk-input-operations
			   with-gtk-grabbed
			   with-interruption
			   without-interruption))
  unspecific)

(define (spawn-edit . args)
  (cond (edwin-editor
	 (error "Edwin is already running."))
	((let ((types (editor-display-types)))
	   (and (pair? types)
		(eq? 'console (display-type/name (car types)))
		(null? (cdr types))))
	 (error "Edwin must run on the console."))
	(else
	 (call-with-current-continuation
	  (lambda (continue)
	    (detach-thread
	     (create-thread continue
	       (lambda ()
		 (set! editor-can-exit? #t)
		 (set! paranoid-exit? #t)
		 (apply edit args)
		 (stop-current-thread)))))))))

(define (update-widgets screen)
  (%trace ";   update-widgets "screen"\n")

  (define-integrable (main)
    (let* ((root (screen-root-window screen)) ;editor-frame
	   (toplevel (gtk-screen-toplevel screen))
	   (top-children (gtk-container-reverse-children toplevel)))
      (update-name screen)
      (if (null? top-children)
	  (let ((top-grid (gtk-grid-new)))
	    (gtk-container-add toplevel top-grid)
	    (%trace ";     -init "root" in "top-grid"\n")
	    (gtk-grid-attach top-grid
			     (re-pack! (editor-frame-typein-window root)#f"--")
			     0 1 1 1)
	    (gtk-grid-attach top-grid
			     (re-pack! (editor-frame-root-window root) #f "--")
			     0 0 1 1)
	    (for-each-text-widget screen update-widget-buffer)
	    (%trace ";     -show-init "toplevel"\n")
	    (gtk-widget-show-all toplevel)
	    (%trace ";   update-widgets init done\n"))
	  (let ((top-grid (car top-children)))
	    (%trace ";     -re-pack "root" into "top-grid"\n")
	    (let ((root-widget
		   (re-pack! (editor-frame-root-window root)
			     (first (gtk-container-reverse-children top-grid))
			     "--")))
	      (if root-widget
		  (begin
		    (%trace ";     -new "root-widget" for "top-grid"\n")
		    (gtk-grid-attach top-grid root-widget 0 0 1 1))))
	    (for-each-text-widget screen update-widget-buffer)
	    (%trace ";     -show-all "toplevel"\n")
	    (gtk-widget-show-all toplevel)
	    (%trace ";   update-widgets done\n")))))

  (define (re-pack! window widget prefix)
    (cond
     ;; Leaf match.
     ((and (buffer-frame? window)
	   (buffer-frame-widget? widget)
	   (let ((text (buffer-frame-widget-text* widget)))
	     (and text
		  (eq? window (text-widget-buffer-frame text))
		  text)))
      => (lambda (text)
	   (%trace ";     "prefix"matched "window" to "
		   widget" (containing "text")\n")
	   (re-size! text window prefix)
	   #f))

     ;; Combo match.
     ((and (combination? window)
	   (if (combination-vertical? window)
	       (gtk-vpaned? widget)
	       (gtk-hpaned? widget)))
      (%trace ";     "prefix"matched "window" to "widget"\n")
      (re-pack-combo! window widget prefix)
      #f)

     (else
      ;; Mismatch.
      (%trace ";     "prefix"mismatched "window" to "widget"\n")
      (if widget
	  (begin
	    (gtk-widget-destroy widget)
	    (%trace ";     "prefix"destroyed "widget"\n")))
      (cond ((buffer-frame? window)
	     (let ((widget (pack-buffer-frame window prefix)))
	       (%trace ";     "prefix"new "widget" for "window"\n")
	       widget))
	    ((combination? window)
	     (let ((widget (pack-combo window prefix)))
	       (%trace ";     "prefix"new "widget" for "window"\n")
	       widget))
	    (else
	     (error "Unexpected Edwin window:" window))))))

    (define (re-pack-combo! combo widget prefix)
      (%trace ";     "prefix"re-pack-combo! "combo" "widget"\n")
      ;; Widget should be a match, orientation-wise.
      (assert (and (combination? combo)
		   (if (combination-vertical? combo)
		       (gtk-vpaned? widget)
		       (gtk-hpaned? widget))))
      (re-pack-combo-child! (combination-child combo) widget
			    (combination-vertical? combo)
			    (string-append "--" prefix))
      #f)

    (define (re-pack-combo-child! child paned vertical? prefix)
      (%trace ";     "prefix"re-pack-combo-child! "child" "paned"\n")
      (let ((next (window-next child)))
	(if (not next)
	    ;; If last, re-pack child as paned's child2.
	    (gtk-paned-pack2-if (re-pack! child (gtk-paned-get-child2 paned)
					  prefix)
				paned prefix)
	    ;; Else as paned's child1.  Init child2 and loop.
	    (begin
	      (gtk-paned-pack1-if (re-pack! child (gtk-paned-get-child1 paned)
					    prefix)
				  paned prefix)
	      (re-pack-combo-child! next
				    (if (window-next next)
					(find/create-paned paned
							   vertical? prefix)
					;; If last, use paned's child2.
					paned)
				    vertical? prefix)))))

    (define (find/create-paned paned vertical? prefix)
      (let ((child2 (gtk-paned-get-child2 paned)))
	(if (if vertical?
		(gtk-vpaned? child2)
		(gtk-hpaned? child2))
	    child2
	    (let ((new (if vertical?
			   (gtk-vpaned-new)
			   (gtk-hpaned-new))))
	      (gtk-paned-pack2-if new paned prefix)
	      new))))

    (define (pack-combo window prefix)
      (%trace ";     "prefix"pack "window"\n")
      (assert (combination? window))
      (let ((new (if (combination-vertical? window)
		     (gtk-vpaned-new)
		     (gtk-hpaned-new))))
	(re-pack-combo-child! (combination-child window) new
			      (combination-vertical? window)
			      (string-append "--" prefix))
	new))

    (define (pack-buffer-frame window prefix)
      (%trace ";     "prefix"pack "window"\n")
      (let ((vgrid (make-buffer-frame-widget))
	    (text (make-text-widget screen
				    (%text-x-size window)
				    (%text-y-size window)))
	    (scroller (gtk-scrolled-view-new))
	    (modeline (and (frame-modeline-inferior window)
			   (make-modeline-widget screen)))
	    (y-step (fix:+ (gtk-screen-line-height screen)
			   (gtk-screen-line-spacing screen)))
	    (x-step (gtk-screen-char-width screen)))
	(set-text-widget-buffer-frame! text window)
	(set-text-widget-modeline! text modeline)
	(set-fix-layout-scroll-step! text x-step y-step)
	;;(gtk-scrolled-window-set-placement scroller 'bottom-right)
	(if (not modeline)
	    ;; No modeline: the window/text-widget should NOT expand.
	    (begin
	      (gtk-scrolled-window-set-policy scroller 'never 'always)
	      (gtk-widget-set-vexpand text #f)
	      (gtk-container-add scroller text)
	      (gtk-container-add vgrid scroller))
	    ;; With modeline: vgrid and scroller SHOULD expand.
	    (begin
	      (gtk-scrolled-window-set-policy scroller 'auto 'always)
	      (gtk-container-add scroller text)
	      (gtk-container-add vgrid scroller)
	      (gtk-container-add vgrid modeline)))
	vgrid))

    (define (re-size! widget window prefix)
      (let ((area (fix-widget-geometry widget))
	    (window-x-size (%text-x-size window))
	    (window-y-size (%text-y-size window)))
	(let ((width (fix-rect-width area))
	      (height (fix-rect-height area)))
	  (if (or (not width) (not height))
	      (%trace ";     "prefix"re-size!: unrealized "widget"\n")
	      (let ((widget-x-size (width->x-size screen width))
		    (widget-y-size (height->y-size screen height)))
		(if (and (fix:= widget-x-size window-x-size)
			 (fix:= widget-y-size window-y-size))
		    (%trace ";     "prefix"re-size!: no change\n")
		    (let ((new-width (x-size->width screen window-x-size))
			  (new-height (y-size->height screen window-y-size)))
		      (%trace ";     "prefix"new natural size for "widget
			      ": from "widget-x-size"x"widget-y-size" "
			      "("width"x"height")"
			      " to "window-x-size"x"window-y-size" "
			      "("new-width"x"new-height")\n")
		      (let unset-pos ((w (gtk-screen-toplevel screen)))
			(if (gtk-container? w)
			    (begin
			      (for-each unset-pos
					(gtk-container-reverse-children w))
			      (if (gtk-paned? w)
				  (gtk-paned-set-position w -1)))))
		      (set-scm-widget-natural-size!
		       widget new-width new-height)
		      (gtk-widget-queue-resize-no-redraw widget))))))))

    (declare (integrate-operator gtk-paned-pack1-if))
    (define (gtk-paned-pack1-if child paned prefix)
      (%trace ";     "prefix"pack1-if "child" "paned"\n")
      (if child
	  (let ((existing (gtk-paned-get-child1 paned)))
	    (if (and existing (not (gtk-widget-destroyed? existing)))
		(begin
		  (%trace ";     "prefix"  replacing "existing"\n")
		  (gtk-widget-destroy existing)))
	    (gtk-paned-pack1 paned child 'resize #f))))

    (declare (integrate-operator gtk-paned-pack2-if))
    (define (gtk-paned-pack2-if child paned prefix)
      (%trace ";     "prefix"pack2-if "child" "paned"\n")
      (if child
	  (let ((existing (gtk-paned-get-child2 paned)))
	    (if (and existing (not (gtk-widget-destroyed? existing)))
		(begin
		  (%trace ";     "prefix"  replacing "existing"\n")
		  (gtk-widget-destroy existing)))
	    (gtk-paned-pack2 paned child 'resize #f))))

    (main))

(define-integrable (typein-widget screen)
  (let* ((top-grid (car (gtk-container-reverse-children
			 (gtk-screen-toplevel screen))))
	 ;; Typein widget is always added first -- last in the reverse list.
	 (typein-frame (last (gtk-container-reverse-children top-grid))))
    (any-child text-widget? typein-frame)))

(define (%children window)
  ;; Produce a list of a combination window's children from left to right
  ;; (or top to bottom).
  (cond ((editor-frame? window)
	 (list (editor-frame-root-window window)
	       (editor-frame-typein-window window)))
	((combination? window)
	 (let loop ((child (combination-child window)))
	   (if child
	       (cons child (loop (window-next child)))
	       '())))
	(else (error "Unexpected Edwin window:" window))))

(define-integrable (%text-x-size window)
  (%window-x-size (frame-text-inferior window)))

(define-integrable (%text-y-size window)
  (%window-y-size (frame-text-inferior window)))

(define-integrable (%widget-x-size window screen)
  (x-size->width screen (%text-x-size window)))

(define-integrable (%widget-y-size window screen)
  (y-size->height screen (%text-y-size window)))

(define (for-each-text-widget screen procedure)
  (every-child (lambda (widget)
		 (and (text-widget? widget)
		      (procedure widget))
		 #t)
	       (gtk-screen-toplevel screen)))

(define (every-text-widget screen predicate)
  ;; Returns #t iff predicate returns #t for every text widget on the
  ;; screen.
  (every-child (lambda (widget)
		 (or (not (text-widget? widget))
		     (predicate widget)))
	       (gtk-screen-toplevel screen)))

(define (any-text-widget container)
  (any-child text-widget? container))

(define (any-child predicate container)
  (let loop ((children (gtk-container-reverse-children container)))
    (cond ((null? children) #f)
	  ((predicate (car children)) (car children))
	  ((gtk-container? (car children))
	   (or (loop (gtk-container-reverse-children (car children)))
	       (loop (cdr children))))
	  (else
	   (loop (cdr children))))))

(define (every-child predicate container)
  (let loop ((children (gtk-container-reverse-children container)))
    (cond ((null? children) #t)
	  ((gtk-container? (car children))
	   (and (loop (gtk-container-reverse-children (car children)))
		(loop (cdr children))))
	  ((predicate (car children)) (loop (cdr children)))
	  (else #f))))

;;; Text and Modeline Widgets

(define-class <edwin-widget>
    (<fix-layout>)

  (screen define standard))

(define-class (<text-widget>
	       (constructor make-text-widget (screen) (x-size y-size)))
    (<edwin-widget>)

  (override-drawing define standard)
  (buffer-drawing define standard initial-value #f)

  ;; Scroll pos for buffer-drawing, saved while override-drawing is up.
  (text-pos define standard initializer (lambda () (cons 0 0)))

  (buffer-frame define standard)
  (modeline define standard initial-value #f)
  (cursor-ink define standard initial-value #f)
  (start-mark define standard initial-value #f))

(define-guarantee text-widget "a <text-widget>")

(define-method initialize-instance ((widget <text-widget>) x-size y-size)
;;;  (%trace ";(initialize-instance <text-widget>) "widget
;;;	  " "x-size" "y-size"\n")
  (let ((screen (edwin-widget-screen widget)))
    (call-next-method widget
		      (x-size->width screen x-size)
		      (y-size->height screen y-size)
		      'white))
  (gtk-widget-set-hexpand widget #t)
  (gtk-widget-set-vexpand widget #t)
  (let ((drawing (make-fix-drawing)))
;;;    (%trace "; drawing: "drawing"\n")
    (let ((ink (make-simple-text-ink)))
      (set-simple-text-ink-text! ink widget "Initial override message.")
      (set-text-ink-color! ink "black")
      (fix-drawing-add-ink! drawing ink)
      (let ((extent (fix-ink-extent ink)))
	(set-fix-drawing-size! drawing
			       (fix-rect-width extent)
			       (fix-rect-height extent))))
    (set-text-widget-override-drawing! widget drawing)
    (set-fix-layout-drawing! widget drawing 0 0))
  (set-fix-widget-focus-change-handler! widget focus-change-handler)
  (set-fix-widget-visibility-notify-handler! widget visibility-notify-handler)
  (set-fix-widget-key-press-handler! widget key-press-handler)
  (set-fix-widget-button-handler! widget 'press button-down-handler)
  widget)

(define-method gtk-widget-destroy-callback ((widget <text-widget>))
  ;; NOTE that this callback can be called before a widget is realized(!).
  (call-next-method widget)
  (let ((cursor (text-widget-cursor-ink widget)))
    (if cursor
	(fix-ink-remove! cursor)))
  (let ((drawing (text-widget-override-drawing widget)))
    (and drawing
	 (let ((ink (car (fix-drawing-display-list drawing))))
	   (and (simple-text-ink? ink)
		(let ((layout (simple-text-ink-pango-layout ink)))
		  (and layout
		       (gobject-unref! layout)))))))
  unspecific)

(define-method fix-widget-realize-callback ((widget <text-widget>))
  (%trace "; (fix-widget-realize-callback <text-widget>) "widget"\n")
  (let ((geometry (fix-widget-geometry widget)))
    (if (or (not (fix-rect-width geometry))
	    (not (fix-rect-height geometry)))
	;; Unfortunately a widget can be realized before it is
	;; allocated a size -- when it is added to a realized
	;; container.  In this case, initialize widget's size to
	;; something reasonable.
	(let ((window (text-widget-buffer-frame widget))
	      (screen (edwin-widget-screen widget)))
	  (%trace "; uninitialized geometry: "geometry"\n")
	  (set-fix-rect-size! geometry
			      (%widget-x-size window screen)
			      (%widget-y-size window screen))
	  (%trace "; initialized geometry: "geometry"\n"))))
  ;; The typein widget is undisturbed by window changes, so grab-focus
  ;; once here.
  (if (not (text-widget-modeline widget))
      (gtk-widget-grab-focus widget))
  (call-next-method widget))

(define-method fix-widget-new-geometry-callback ((widget <text-widget>))
  (%trace "; (fix-widget-new-geometry-callback <text-widget>) "widget"\n")
  (call-next-method widget)
  (let ((screen (edwin-widget-screen widget)))
    (queue-input-event
     screen
     (make-input-event
      'set-window-size
      (lambda (widget)
	(%trace ";  input event: set-window-size "widget"\n")
	(let ((geometry (fix-widget-geometry widget))
	      (window (text-widget-buffer-frame widget)))
	  (let ((widget-x-size (width->x-size screen (fix-rect-width geometry)))
		(widget-y-size (height->y-size screen (fix-rect-height geometry)))
		(window-x-size (%text-x-size window))
		(window-y-size (%text-y-size window)))
	    (%trace ";    "widget": "geometry"\n")
	    (%trace ";    "window": "window-x-size"x"window-y-size"\n")
	    (if (not (and (fix:= widget-x-size window-x-size)
			  (fix:= widget-y-size window-y-size)))
		(update-sizes screen)))))
      widget))))

(define (update-sizes screen)
  ;; The underhanded way to adjust window sizes.  This procedure does
  ;; not use the :set-size! method, which presumably adjusts the
  ;; widget sizes.  It does the "opposite".  It leaves the widgets
  ;; alone and adjusts Edwin's window and screen sizes (using %
  ;; operators).

  (define (%set-size! screen window prefix)
    (cond
     ((buffer-frame? window)
      (let ((widget (window-text-widget* window)))
	(if widget
	    (let* ((view (fix-layout-view widget))
		   (width (fix-rect-width view))
		   (height (fix-rect-height view))
		   (x-size (width->x-size screen width))
		   (y-size (height->y-size screen height))
		   (x-size* (if (window-has-right-neighbor? window)
				(fix:1+ x-size) x-size))
		   (y-size* (if (frame-modeline-inferior window)
				(fix:1+ y-size) y-size))
		   (text (frame-text-inferior window)))

	      (%trace ";   "prefix""text": "x-size"x"y-size" "view"\n")
	      (%set-window-x-size! text x-size)
	      (%set-window-y-size! text y-size)
	      (%trace ";   "prefix""window": "x-size*"x"y-size*"\n")
	      (%set-window-x-size! window x-size*)
	      (%set-window-y-size! window y-size*))
	    (%trace ";   "prefix""window": no widget\n"))))

     ((or (combination? window)
	  (editor-frame? window))
      (let ((total-x-size #f)
	    (total-y-size #f)
	    (vertical? (or (editor-frame? window)
			   (combination-vertical? window))))
	(for-each
	  (lambda (inferior)
	    (let ((child (inferior-window inferior)))
	      (%set-size! screen child (string-append prefix "--"))
	      (if vertical?
		  (let ((x-size (%window-x-size child))
			(y-size (%window-y-size child)))
		    (set! total-x-size
			  (if (not total-x-size)
			      x-size
			      (fix:max x-size total-x-size)))
		    (set! total-y-size
			  (if (not total-y-size)
			      y-size
			      (fix:+ total-y-size y-size))))
		  (let ((x-size (%window-x-size child))
			(y-size (%window-y-size child)))
		    (set! total-y-size
			  (if (not total-y-size)
			      y-size
			      (fix:max y-size total-y-size)))
		    (set! total-x-size
			  (if (not total-x-size)
			      x-size
			      (fix:+ total-x-size x-size)))))))
	  (window-inferiors window))
	(%trace ";   "prefix""window": "total-x-size"x"total-y-size"\n")
	(%set-window-x-size! window total-x-size)
	(%set-window-y-size! window total-y-size)))

     (else
      (%trace ";   "prefix""window": unexpected type\n"))))

  (define (%set-starts! windows parent prefix x y)
    (if (pair? windows)
	(let* ((window (car windows))
	       (inferior (find-inferior (window-inferiors parent) window)))
	  (%trace ";   "prefix""window" start: "x"x"y
		  " (was "
		  (inferior-x-start inferior)"x"(inferior-y-start inferior)
		  ")\n")
	  (%set-inferior-start! inferior x y)
	  (if (or (editor-frame? window)
		  (combination? window))
	      (%set-starts! (%children window) window
			    (string-append prefix "--")
			    x y))
	  (if (or (editor-frame? parent)
		  (and (combination? parent)
		       (combination-vertical? parent)))
	      (%set-starts! (cdr windows) parent prefix
			    x
			    (fix:+ y (%window-y-size window)))
	      (%set-starts! (cdr windows) parent prefix
			    (fix:+ x (%window-x-size window))
			    y)))))

  (%trace "; update-sizes "screen"\n")
  (let ((root (screen-root-window screen)))
    (let ((x-size (%window-x-size root))
	  (y-size (%window-y-size root)))
      (%trace ";   initial root size: "x-size"x"y-size"\n"))
    (%set-size! screen root "--")
    (let ((x-size (%window-x-size root))
	  (y-size (%window-y-size root)))
      (%trace ";   screen: "x-size"x"y-size"\n")
      (set-screen-x-size! screen x-size)
      (set-screen-y-size! screen y-size))
    (%set-starts! (%children root) root "--" 0 0)))

(define-integrable (editor-frame? object)
  (object-of-class? editor-frame object))

(define-method fix-layout-adjustment-parameters ((widget <text-widget>)
						 view-size extent-size
						 extent-low receiver)
  (declare (ignore widget))
  (if (fix:< view-size extent-size)
      ;; Drawing is larger than viewport: thumb (page) is viewport.
      (begin
	(%trace ";  large-drawing:"extent-size" view:"view-size"\n")
	(receiver extent-low view-size))
      ;; Viewport is larger than drawing: thumb (page) is 0.
      (begin
	(%trace ";  drawing:"extent-size" large-view:"view-size"\n")
	;; When page-size = extent-size, high will be extent-low [see
	;; adjust-adjustment].  Thus high and low are extent-low, and
	;; value will be clamped to extent-low.
	(receiver extent-low extent-size))))

(define-class (<modeline-widget> (constructor make-modeline-widget (screen)))
    (<edwin-widget>))

(define-method initialize-instance ((widget <modeline-widget>))
;;;  (%trace ";(initialize-instance <modeline-widget>) "widget"\n")
  (let ((screen (edwin-widget-screen widget)))
    (call-next-method widget 0 (y-size->height screen 1) 'black))
  (gtk-widget-set-hexpand widget #t)
  (gtk-widget-set-vexpand widget #f)
  (gtk-widget-set-name widget "modeline")
  (let ((drawing (make-fix-drawing)))
;;;    (%trace "; drawing: "drawing"\n")
    (set-fix-layout-drawing! widget drawing 0 0))
  (set-fix-widget-focus-change-handler! widget focus-change-handler)
  (set-fix-widget-visibility-notify-handler! widget visibility-notify-handler)
  (set-fix-widget-key-press-handler! widget key-press-handler)
  widget)

(define-method fix-widget-realize-callback ((widget <modeline-widget>))
  (%trace "; (fix-widget-realize-callback <modeline-widget>) "widget"\n")
  (let ((screen (edwin-widget-screen widget))
	(geometry (fix-widget-geometry widget)))
    (if (or (not (fix-rect-width geometry))
	    (not (fix-rect-height geometry)))
	;; Unfortunately a widget can be realized before it is
	;; allocated a size -- when it is added to a realized
	;; container.  In this case, initialize widget's size to
	;; something reasonable.
	(begin
	  (%trace "; uninitialized geometry: "geometry"\n")
	  (set-fix-rect-size! geometry 0 (y-size->height screen 1))
	  (%trace "; initialized geometry: "geometry"\n")))
    (call-next-method widget)
    (let ((ink (make-simple-text-ink))
	  (drawing (fix-layout-drawing widget)))
      (set-simple-text-ink-font! ink (gtk-screen-font screen))
      (set-simple-text-ink-text!
       ink widget "--------Initial mode line.--------------------------------")
      (set-text-ink-color! ink "white")
      (fix-drawing-add-ink! drawing ink)
      (let ((extent (fix-ink-extent ink)))
	(set-fix-drawing-size! drawing
			       (fix-rect-width extent)
			       (fix-rect-height extent))))))

(define-class (<buffer-frame-widget> (constructor ()))
    (<gtk-grid>)

  ;; This one just "marks" a gtk-container as the type that holds a
  ;; text-widget, modeline and button bars together.  If the frame has
  ;; no modeline nor buttons, a lone scroller STILL gets wrapped.
  )

(define-method initialize-instance ((widget <buffer-frame-widget>))
;;;  (%trace ";(initialize-instance <buffer-frame-widget>) "widget"\n")
  (call-next-method widget)
  (gtk-orientable-set-orientation widget 'vertical))

;; Assume there is one text-widget in a buffer-frame-widget.
(define-integrable buffer-frame-widget-text* any-text-widget)

(define-class (<gtk-hpaned> (constructor ()))
    (<gtk-paned-view>))

(define (gtk-hpaned-new)
  (let ((widget (make-gtk-hpaned)))
    (gtk-paned-view-init widget 'horizontal)
    (gtk-widget-set-hexpand widget #t)
    widget))

(define-class (<gtk-vpaned> (constructor ()))
    (<gtk-paned-view>))

(define (gtk-vpaned-new)
  (let ((widget (make-gtk-vpaned)))
    (gtk-paned-view-init widget 'vertical)
    (gtk-widget-set-vexpand widget #t)
    widget))

;;; Incremental Redisplay

;; Drawing a Buffer
;;
;; At its simplest, drawing a buffer is a process of searching for
;; the "lines" between newlines and creating a <line-ink> for
;; each.  The <line-ink>s are sized -- layed out in a PangoLayout
;; -- and arranged vertically against the left margin.  Each line-ink
;; remembers the start and end indices of a line in a buffer and the
;; bounding box of the laid-up line/paragraph, and not much else.
;;
;; The incremental version of this process updates an existing column
;; of <line-ink>s after the buffer has changed.  It skips
;; unchanged lines at the top, and re-lays out lines in the change
;; region.  Depending on the newlines in the region, it may re-use
;; lines, create more, or erase some.  Lines below the region are
;; textually unchanged, and do not have to be re-layed out by Pango,
;; though they may need to be moved to accommodate insertions and
;; deletions above them.
;; 
;; <line-ink>s are text-inks, but not simple-text-inks.  The latter
;; keep a PangoLayout around to service expose events.  A drawing of a
;; large buffer, with thousands of lines, if drawn with
;; simple-text-inks, would allocate thousands of PangoLayouts, each
;; with an image of a line.  The images alone would occupy more bytes
;; than the buffer text.
;;
;; To lighten the footprint of a large buffer drawing, line-inks do
;; not hold a PangoLayout, but create one on demand using the buffer
;; text.  They cache the created PangoLayout, and steal existing
;; PangoLayouts from line-inks that are off-screen.  The caching
;; allows most expose events to find exposed line-inks ready with a
;; PangoLayout to paint.  As lines scroll into view, new PangoLayouts
;; are allocated (or stolen), and the buffer text is re-imaged,
;; styled, and re-layed-out just as when originally drawn.
;;
;; Edwin now runs with the glib lock, giving it up only to read input.
;; Thus the expose event handler should not run until the end of
;; redisplay when all drawings are up-to-date and there are no buffer
;; changed regions, allowing an exposed line ink to re-construct its
;; PangoLayout from unchanged buffer text.

(define-method update-screen! ((screen <gtk-screen>) display-style)
  (%trace "; (update-screen! <gtk-screen>) "screen" "display-style"\n")
  (with-screen-in-update
   screen
   (lambda ()
     (cond
      ((display-style/no-screen-output? display-style)
       (invalidate-all-drawings! screen)
       (%trace "; (update-screen! <gtk-screen>) done: no-output\n")
       'no-output)
      ((not (memq (screen-visibility screen) '(visible partially-obscured)))
       (let ((visibility (screen-visibility screen)))
	 (if (not (eq? visibility 'deleted))
	     (update-name screen))
	 (invalidate-all-drawings! screen)
	 (%trace "; (update-screen! <gtk-screen>) done: "visibility"\n")
	 visibility))
      (else
       (update-widgets screen)
       (%trace ";   update drawings\n")
       (if (every (lambda (entry) (update-drawing screen (cdr entry)))
		  (gtk-screen-drawings screen))
	   (begin
	     (%trace ";   update windows\n")
	     (for-each-text-widget screen update-window)
	     (if (display-style/discard-screen-contents? display-style)
		 (for-each-text-widget screen gtk-widget-queue-draw))
	     (update-blinking screen)
	     (%trace "; (update-screen! <gtk-screen>) done: finished\n")
	     #t)
	   (begin
	     (%trace "; (update-screen! <gtk-screen>) done: halted\n")
	     #f)))))))

(define (invalidate-all-drawings! screen)
  (%trace ";   invalidated all drawings\n")
  (for-each (lambda (entry)
	      (set-buffer-drawing-valid?! (cdr entry) #f))
	    (gtk-screen-drawings screen)))

(declare (integrate-operator with-screen-in-update))
(define (with-screen-in-update screen thunk)
  (assert-glib-locked 'with-screen-in-update)
  (if (screen-in-update? screen)
      (error "Recursive update:" screen))
  (set-screen-in-update?! screen #t)
  (let ((v (thunk)))
    (set-screen-in-update?! screen #f)
    v))

(define (update-blinking screen)
  ;; Sometimes called by a callback (i.e. without-interrupts).  Frobs
  ;; JUST the canvas (else must queue an editor input event.)
  (%trace ";   update blinking "screen"\n")
  (if (not (gtk-screen-in-focus? screen))
      (begin
	(%trace ";     not the focus\n")
	(blink! screen #f))
      (let ((window (screen-cursor-window screen)))
	(if (not window)
	    (begin
	      (%trace ";     no cursor window\n")
	      (blink! screen #f))
	    (let ((widget (window-text-widget* window)))
	      (%trace ";     cursor window: "window" "widget"\n")
	      (guarantee-text-widget widget 'update-blinking)
	      (let ((cursor (text-widget-cursor-ink widget)))
		(if (not cursor)
		    (begin
		      (%trace ";     no cursor yet\n")
		      (blink! screen #f))
		    (begin
		      (%trace ";     enabling "cursor"\n")
		      (visible! cursor #t)
		      (blink! screen cursor)))))))))

(define-method update-screen-window!
    ((screen <gtk-screen>) window display-style)
  (%trace "; (update-screen-window! <gtk-screen>) "screen" "window"\n")
  (let ((v (update-screens! display-style)))
    (%trace "; (update-screen-window! <gtk-screen>) "screen" "window" => "v"\n")
    v))

(define (update-widget-buffer widget)
  (%trace ";     update-widget-buffer "widget"\n")
  (let ((screen (edwin-widget-screen widget))
	(window (text-widget-buffer-frame widget)))
	    
    (define-integrable (main)
      (let* ((new-buffer (window-buffer window))
	     (old-drawing (text-widget-buffer-drawing widget))
	     (old-buffer (and old-drawing
			      (buffer-drawing-buffer old-drawing))))
	(if (and old-buffer (eq? new-buffer old-buffer)
		 old-drawing (drawing-match? old-drawing))
	    (%trace ";\tno change\n")
	    (let ((new-drawing (find/create-drawing widget)))
	      (%trace ";\tnew/old buffer: "new-buffer
		      "/"old-buffer" ("old-drawing")\n")
	      (set-text-widget-buffer-drawing! widget new-drawing)
	      (re-cursor widget new-drawing)
	      (if (not (eq? (fix-layout-drawing widget)
			    (text-widget-override-drawing widget)))
		  (set-fix-layout-drawing! widget new-drawing 0 0))))))

    (define (re-cursor widget drawing)
      ;; Re-set text-widget-cursor-ink per new buffer in drawing.
      (%trace ";\tre-cursor "widget" "drawing"\n")
      (let ((cursor (text-widget-cursor-ink widget))
	    (modeline (text-widget-modeline widget)))
	(cond ((not cursor)
	       (let ((new (make-cursor-ink))
		     (width (fix:quotient (gtk-screen-char-width screen) 2))
		     (height (gtk-screen-line-height screen))
		     (space (gtk-screen-line-spacing screen))
		     (widgets (list widget)))
		 (%trace ";\t  new "new" for new "widget"\n")
		 (set-rectangle-ink! new 0 space width height)
		 (set-cursor-ink-widget-list! new widgets)
		 (if (not modeline)
		     (begin
		       (set-fix-ink-widgets! new '())
		       (set-cursor-ink-visible?! new #f))
		     (begin
		       (set-fix-ink-widgets! new widgets)))
		 (set-text-widget-cursor-ink! widget new)
		 (fix-drawing-add-ink! drawing new 'bottom)))
	      ((not (eq? drawing (fix-ink-drawing cursor)))
	       (%trace ";\t  moving "cursor" to new "drawing"\n")
	       (fix-ink-remove! cursor)
	       (set-rectangle-ink-position! cursor
					    0 (gtk-screen-line-spacing screen))
	       (fix-drawing-add-ink! drawing cursor 'bottom))
	      (else
	       (%trace ";\t  no change\n")))))

    (define (find/create-drawing widget)
      (%trace ";\tfind/create-drawing for "widget" ("window")\n")
      (let ((buffer (window-buffer window))
	    (drawings (gtk-screen-drawings screen)))
	(or
	 (cdr* (find (lambda (buffer.drawing)
		       (and (eq? (car buffer.drawing) buffer)
			    (drawing-match? (cdr buffer.drawing))))
		     drawings))
	 (let* ((bufwin (frame-text-inferior window))
		(new (make-buffer-drawing
		      buffer
		      (%window-tab-width bufwin)
		      (%window-char-image-strings bufwin))))
	   (%trace ";\t  new buffer drawing: "new" "buffer
		   " "window" "widget"\n")
	   (set-gtk-screen-drawings! screen (cons (cons buffer new) drawings))
	   new))))

    (define (drawing-match? drawing)
      ;; #t iff nothing has changed, in terms of drawing style
      ;; parameters, between window and drawing.
      (let ((bufwin (frame-text-inferior window)))
	(and (fix:= (%window-tab-width bufwin)
		    (buffer-drawing-tab-width drawing))
	     (eq? (%window-char-image-strings bufwin)
		  (buffer-drawing-char-image-strings drawing)))))

    (main)))

(define (update-window widget)
  (%trace ";     update-window "widget"\n")
  (let ((window (text-widget-buffer-frame widget)))
    (let ((message (window-override-message window))
	  (drawing (fix-layout-drawing widget)) ; current drawing: either...
	  (override (text-widget-override-drawing widget)) ; this...
	  (text (text-widget-buffer-drawing widget)))	   ; or this.
      (guarantee-fix-drawing drawing 'update-window)
      (guarantee-fix-drawing override 'update-window)
      (guarantee-fix-drawing text 'update-window)
      (if message
	  (begin
	    ;; ReDisplay message in override.
	    (let* ((text-ink (car (fix-drawing-display-list override))))
	      (set-simple-text-ink-text! text-ink widget message)
	      (let ((e (fix-ink-extent text-ink)))
		(set-fix-drawing-size! override
				       (fix-rect-width e)
				       (fix-rect-height e))))
	    (if (not (eq? override drawing))
		(let ((saved-pos (text-widget-text-pos widget))
		      (view (fix-layout-view widget)))
		  (set-car! saved-pos (fix-rect-x view))
		  (set-cdr! saved-pos (fix-rect-y view))
		  (%trace ";\tsaving text position "saved-pos"\n")
		  (set-fix-layout-drawing! widget override 0 0))
		(%trace ";\toverride still up\n")))
	  (begin
	    ;; ReDisplay text, and scroll to cursor.
	    (if (not (eq? text drawing))
		(let ((saved-pos (text-widget-text-pos widget)))
		  (%trace ";\trestoring "text" to "saved-pos"\n")
		  (set-fix-layout-drawing! widget text
					   (car saved-pos) (cdr saved-pos)))
		(%trace ";\ttext still up\n"))
	    (update-cursor widget)
	    (let ((extent (fix-ink-extent (text-widget-cursor-ink widget))))
	      (%trace ";\tscrolling to "extent"\n")
	      (fix-layout-scroll-nw! widget extent)
	      (%trace ";\tview: "(fix-layout-view widget)"\n"))
	    (update-modeline widget))))))

;; This variable caches a modeline image buffer.  A modeline update
;; hacks this buffer, then compares it to the string in the simple-
;; text-ink.  This avoids much consing and widget damage.  The Edwin
;; thread should be the only thread accessing this resource.
(define modeline-image "")

(define (update-modeline widget)
  ;; Must be last in the update process.  Some of its state depends on
  ;; the final scroll position!
  (%trace ";     update-modeline "widget"\n")
  (let* ((window (text-widget-buffer-frame widget))
	 ;; Add a few columns so the text runs past scrollbars and
	 ;; whatnot, off the right side of the widget.
	 (x-size (+ 5 (%window-x-size window))))
    (if widget
	(let ((modeline (text-widget-modeline widget)))
	  (if modeline
	      (begin
		(let ((maxlen (string-maximum-length modeline-image)))
		  (if (> x-size maxlen)
		      (set! modeline-image (string-allocate x-size))
		      (set-string-length! modeline-image maxlen)))
		(modeline-string! window modeline-image 0 x-size)
		(set-string-length! modeline-image x-size)
		(let* ((drawing (fix-layout-drawing modeline))
		       (inks (fix-drawing-display-list drawing))
		       (ink (cond ((null? inks)
				   (let ((i (make-simple-text-ink)))
				     (fix-drawing-add-ink! drawing i)
				     i))
				  ((simple-text-ink? (car inks)) (car inks))
				  (else (error "bogus modeline drawing"))))
		       (old (simple-text-ink-text ink)))
		  (if (not (and old (string=? old modeline-image)))
		      (let ((copy (string-copy modeline-image)))
			(set-simple-text-ink-text! ink widget copy)
			;; Ensure that text-ink is wider than widget???
			(%trace ";\tupdated "modeline": \""copy"\"\n"))
		      (%trace ";\tunchanged "modeline"\n"))))
	      (%trace ";\tno modeline\n")))
	(%trace ";\tno widget!\n")))
  (%trace ";     update-modeline done\n"))

(define (update-name screen)
  (let ((name (frame-name screen))
	(name* (gtk-screen-name screen)))
    (if (and name (or (not name*) (not (string=? name name*))))
	(begin
	  (set-gtk-screen-name! screen name)
	  (gtk-window-set-title (gtk-screen-toplevel screen) name)))))

(define (frame-name screen)
  (let* ((window
	  (if (and (eq? screen (selected-screen)) (within-typein-edit?))
	      (typein-edit-other-window)
	      (screen-selected-window screen)))
	 (buffer (window-buffer window))
	 (format (ref-variable frame-name-format buffer)))
    (and format
	 (string-trim-right
	  (format-modeline-string
	   window format (ref-variable frame-name-length buffer))))))

(define (update-drawing screen drawing)
  ;; Redraw a buffer-drawing.
  (%trace ";     update-drawing "screen" "drawing"\n")

  ;; This is the traditional Emacs layout, in a fixed-width font, with
  ;; 2 and 4 character depictions of many characters (e.g. ^@ and
  ;; \200).

  ;; Line wrapping is not currently supported.

  ;; Consider first a diagram of our buffer:
  ;; 
  ;;     unchanged prefix
  ;;     change-region
  ;;     unchanged suffix
  ;; 
  ;; and the process of redrawing it:
  ;; 
  ;;     Skip through prefix, to a line needing updating -- a line
  ;;     stretching into the change region.  There may be no such line
  ;;     if there is no next line, or the next line does not need
  ;;     updating -- lies beyond the change region.
  ;; 
  ;;     Steal this line-needing-updating (if any); lay it out again;
  ;;     move/re-size it.  Steal it AND the next... until the last
  ;;     stolen line reaches beyond the change region, or there are no
  ;;     more lines-needing-updating to steal.  Remove any remaining
  ;;     lines-needing-updating.  If the last stolen line did NOT
  ;;     reach beyond the change region (nor hit the buffer's end),
  ;;     add lines until the last added line does.  The last stolen or
  ;;     added line should MEET the next line, a line NOT needing
  ;;     updating (if any, else the buffer's end).
  ;;
  ;;     Move the remaining lines -- those entirely in the suffix (if
  ;;     any).  Note that if the first remaining line does not need to
  ;;     move, neither do the rest.
  ;; 
  ;; Now consider display-start/end:
  ;; 
  ;;     Remove lines starting before display-start.
  ;;     Steal/add lines until they match the prefix (or perhaps the
  ;;     suffix, OR the display-end).
  ;; 
  ;;     As before, skip through the prefix, except that these
  ;;     "unchanged" lines might have to move.
  ;; 
  ;;     As before, steal/add changed lines until they reach (and
  ;;     meet!) lines in the suffix.
  ;; 
  ;;     As before, move lines in the suffix as necessary, except do
  ;;     not bother with lines reaching beyond display-end.
  ;; 
  ;;     Remove lines extending beyond display-end.  Add lines until
  ;;     they hit display-end.

  (let* ((line-height (gtk-screen-line-height screen))
	 (line-spacing (gtk-screen-line-spacing screen))
	 (drawing-extent #f)	       ;set when an ink extent is known
	 (pango-layout #f)	       ;set when a pango-layout is allocated
	 (buffer (buffer-drawing-buffer drawing))
	 (group (buffer-group buffer))
	 (display-start (group-display-start group))
	 (display-end (group-display-end group))
	 (change-start-index (if (buffer-drawing-valid? drawing)
				 (group-start-changes-index group)
				 (mark-index display-start)))
	 (change-end-index (if (buffer-drawing-valid? drawing)
			       (group-end-changes-index group)
			       (mark-index display-end))))

    (define-syntax %trace3
      (syntax-rules ()
	((_ args ...) (if %trace-redraw?
			  (apply outf-error (%trace-simplify args ...))))))

    (define-integrable (main)
      (%trace3 ";\tdrawing/buffer ticks:"
	       " "(buffer-drawing-modified-tick drawing)
	       "/"(group-modified-tick group)"\n"
	       ";\tchange/display regions:"
	       " "change-start-index"-"change-end-index
	       "/"display-start"-"display-end"\n")
      (init-start/end)
      (cond
       ((no-display-changes?)
	(%trace ";\tno changes\n")
	#t)
       (else
	(let ((finished?

	       (redraw-start
		(next-lines (fix-drawing-display-list drawing))
		display-start 1 line-spacing
		(lambda (lines start num y)

		  (redraw-prefix
		   lines start num y
		   (lambda (lines start num y)

		     (redraw-changed
		      lines start num y
		      (lambda (lines start num y)

			(redraw-suffix
			 lines start num y

			 redraw-end)))))))))

	  (if finished?
	      (begin
		(set-size)
		(move-mark-to! (buffer-drawing-display-start drawing)
			       display-start)
		(move-mark-to! (buffer-drawing-display-end drawing)
			       display-end)
		(set-buffer-drawing-modified-tick!
		 drawing (group-modified-tick group))
		(set-buffer-drawing-valid?! drawing #t)))

	  (if pango-layout (gobject-unref! pango-layout))
	  finished?))))

    (define-integrable (init-start/end)
      (if (not (buffer-drawing-display-start drawing))
	  (begin
	    (set-buffer-drawing-display-start! drawing
					       (mark-permanent-copy
						display-start))
	    (set-buffer-drawing-display-end! drawing
					     (mark-permanent-copy
					      display-end)))))
    (define-integrable (set-size)
      (if drawing-extent
	  (let ((width+
		 (fix:+ (fix-rect-max-x drawing-extent)
			(gtk-screen-char-width screen)))
		(height+
		 (fix:+ (fix-rect-max-y drawing-extent)
			(if (final-newline? group)
			    (fix:+ line-spacing
				   (fix:+ line-height
					  line-spacing))
			    line-spacing))))
	    (fix-rect-union! drawing-extent (make-fix-rect 0 0 width+ height+))
	    (if (not (and (fix:= (fix-rect-min-x drawing-extent) 0)
			  (fix:= (fix-rect-min-y drawing-extent) 0)))
		(%trace "; Warning: drawing min x,y"
		       " = "(fix-rect-min-x drawing-extent)
		       ","(fix-rect-min-y drawing-extent)"!\n"))
	    (set-fix-drawing-size! drawing
				   (fix-rect-max-x drawing-extent)
				   (fix-rect-max-y drawing-extent)))
	  (set-fix-drawing-size! drawing 0 0))
      (%trace ";\tnew drawing extent: "(fix-drawing-extent drawing)"\n"))

    (define (redraw-start lines start num y receiver)
      (%trace3 ";         redraw-start "lines" "start" "num" "y"\n")
      (let ((old-start (and (more-lines? lines)
			    (%unchanged? (car lines)
					 change-start-index change-end-index)
			    (line-ink-start (car lines)))))
	(cond ((not old-start)
	       (%trace3 ";           hit changed "(and(not(null? lines))(car lines))"\n")
	       (receiver lines start num y))
	      ((mark= start old-start)
	       (%trace3 ";           matched "(car lines)"\n")
	       (receiver lines start num y))
	      ((mark< start old-start)
	       (let ((new (add-line start num y lines)))
		 (%trace3 ";           added "new"\n")
		 (redraw-start lines (next-start new)
			       (next-num num) (next-y new) receiver)))
	      ((mark< old-start start)	;uncommon
	       (redraw-start (remove-lines-before lines start)
			     start num y receiver))
	      (else (%trace3 "; Unreachable?!\n")))))

    (define (redraw-prefix lines start num y receiver)
      (%trace3 ";        redraw-prefix "lines" "start" "num" "y"\n")
      (cond (((editor-halt-update? current-editor))
	     (%trace3 ";           halt redraw!\n")
	     #f)
	    ((not (more-lines? lines))
	     (%trace3 ";           no more lines\n")
	     (receiver lines start num y))
	    ((and (%unchanged? (car lines) change-start-index change-end-index)
		  (mark<= (line-ink-end (car lines)) display-end))
	     (let ((next-y (move-line! (car lines) start num y)))
	       (%trace3 ";           prefix "(car lines)"\n")
	       (redraw-prefix (next-lines (cdr lines))
			      (next-start (car lines))
			      (next-num num)
			      next-y
			      receiver)))
	    (else
	     (%trace3 ";           not prefix "(car lines)"\n")
	     (receiver lines start num y))))

    (define (redraw-changed lines start num y receiver)
      (%trace3 ";         redraw-changed "lines" "start" "num" "y"\n")
      (if (not change-start-index)
	  (begin
	    (%trace3 ";           no change region\n")
	    (receiver lines start num y))
	  (steal-changed
	   lines start num y
	   (lambda (lines start num y)
	     (remove-changed
	      lines start num y
	      (lambda (lines start num y)
		(add-changed
		 lines start num y
		 (lambda (lines start num y)
		   (receiver lines start num y)))))))))

    (define (steal-changed lines start num y receiver)
      (%trace3 ";           steal-changed "lines" "start" "num" "y"\n")
      (cond (((editor-halt-update? current-editor))
	     (%trace3 ";             halt redraw!\n")
	     #f)
	    ((not (more-lines? lines))
	     (%trace3 ";             no more lines\n")
	     (receiver lines start num y))
	    ((mark<= display-end start)
	     (%trace3 ";             hit end at "start" with "lines"\n")
	     (receiver lines start num y))
	    ((%unchanged? (car lines) change-start-index change-end-index)
	     (%trace3 ";             unchanged "(car lines)"\n")
	     (receiver lines start num y))
	    ((fix:< change-start-index (mark-index start))
	     (%trace3 ";             beyond changes at "start"\n")
	     (receiver lines start num y))
	    (else
	     (steal-line! (car lines) start num y)
	     (%trace3 ";             stole line "(car lines)"\n")
	     (let* ((line (car lines))
		    (next-start (next-start line))
		    (next-lines (next-lines (cdr lines))))
	       (steal-changed (remove-lines-before next-lines next-start)
			      next-start (next-num num) (next-y line)
			      receiver)))))

    (define (remove-changed lines start num y receiver)
      (%trace3 ";           remove-changed "lines" "start" "num" "y"\n")
      (cond (((editor-halt-update? current-editor))
	     (%trace3 ";             halt redraw!\n")
	     #f)
	    ((not (more-lines? lines))
	     (%trace3 ";             no more lines\n")
	     (receiver lines start num y))
	    ((%unchanged? (car lines) change-start-index change-end-index)
	     (%trace3 ";             unchanged "(car lines)"\n")
	     (receiver lines start num y))
	    (else
	     (remove-line (car lines))
	     (remove-changed (next-lines (cdr lines))
			     start num y receiver))))

    (define (add-changed lines start num y receiver)
      (%trace3 ";           add-changed "lines" "start" "num" "y"\n")
      (cond (((editor-halt-update? current-editor))
	     (%trace3 ";           halt redraw!\n")
	     #f)
	    ((mark<= display-end start)
	     (%trace3 ";           hit end at "start" with "lines"\n")
	     (receiver lines start num y))
	    ((fix:<= (mark-index start) change-end-index)
	     (let* ((new (add-line start num y lines))
		    (new-start (next-start new)))
	       (%trace3 ";           added "new"\n")
	       (add-changed (remove-lines-before lines new-start)
			    new-start (next-num num) (next-y new) receiver)))
	    (else
	     (%trace3 ";           beyond change at "start"\n")
	     (receiver lines start num y))))

    (define (redraw-suffix lines start num y receiver)
      (%trace3 ";         redraw-suffix "lines" "start" "num" "y"\n")
      (cond (((editor-halt-update? current-editor))
	     (%trace3 ";           halt redraw!\n")
	     #f)
	    ((not (more-lines? lines))
	     (%trace3 ";           no more lines\n")
	     (receiver lines start num y))
	    ((mark= display-end start)
	     (%trace3 ";           at end "(car lines)"\n")
	     (receiver lines start num y))
	    ((mark< display-end start)
	     (%trace3 ";           beyond end "(car lines)"\n")
	     (receiver lines start num y))
	    (else
	     (let ((next-y (move-line! (car lines) start num y)))
	       (%trace3 ";           suffix "(car lines)"\n")
	       (redraw-suffix (next-lines (cdr lines))
			      (next-start (car lines))
			      (next-num num)
			      next-y
			      receiver)))))

    (define (redraw-end lines start num y)
      (%trace3 ";         redraw-end "lines" "start" "num" "y"\n")
      (cond (((editor-halt-update? current-editor))
	     (%trace3 ";           halt redraw!\n")
	     #f)
	    ((mark= start display-end)	;common
	     (%trace3 ";           clipping "lines"\n")
	     (remove-lines lines)
	     #t)
	    ((and (mark< start display-end)
		  (more-lines? lines))
	     (steal-line! (car lines) start num y)
	     (%trace3 ";           stole line "(car lines)"\n")
	     (redraw-end (next-lines (cdr lines))
			 (next-start (car lines))
			 (next-num num)
			 (next-y (car lines))))
	    ((mark< start display-end)	;no more lines to steal
	     (let ((new (add-line start num y '())))
	       (%trace3 ";           added "new"\n")
	       (redraw-end lines
			   (next-start new)
			   (next-num num)
			   (next-y new))))
	    (else
	     ;; (mark< display-end start)
	     (%trace3 "; Warning: last line (before "(car lines)")"
		    " ended beyond display-end!\n")
	     (remove-lines lines))))

    ;; If all inks in the drawing go through next-lines,
    ;; move-line!, add-line or steal-line!, then all of their extents
    ;; can be unioned to get the extent encompassing them all.
    (define (union-ink! ink)
      (union-extent! (fix-ink-extent ink)))

    (define (union-extent! extent)
      (if (not drawing-extent)
	  (set! drawing-extent (copy-fix-rect extent))
	  (fix-rect-union! drawing-extent extent)))

    ;; Keeps the next line to redraw on the front, skipping inks like
    ;; cursors, selection boxes, embedded images/widgets/whatnot.
    (define (next-lines inks)
      (cond ((null? inks) '())
	    ((line-ink? (car inks)) inks)
	    ((cursor-ink? (car inks))
	     ;; Punt cursor extents.  They often move around at the last
	     ;; moment. :-)
	     (next-lines (cdr inks)))
	    ((fix-ink? (car inks))
	     (union-ink! (car inks))
	     (next-lines (cdr inks)))
	    (else
	     (%trace3 "; Warning: bogus "(car inks)" in "drawing"\n")
	     (next-lines (cdr inks)))))

    (define-integrable next-start line-ink-end)

    (define-integrable next-num fix:1+)

    (define (next-y line)
      (next-y-extent (fix-ink-extent line)))

    (define (next-y-extent extent)
      (fix:+ (fix-rect-max-y extent) line-spacing))

    (define (more-lines? inks)
      (and (not (null? inks))
	   (line-ink? (car inks))))

    (define (remove-lines-before lines start)
      ;; Used to clear off (erase!) lines that have been run over by
      ;; newly added (stolen) lines.
      (%trace3 ";           remove-lines-before "start" "lines"\n")
      (cond ((null? lines) '())
	    ((mark< (line-ink-start (car lines)) start)
	     (remove-line (car lines))
	     (remove-lines-before (next-lines (cdr lines)) start))
	    (else lines)))

    (define (remove-lines lines)
      ;; Used to clear off lines that hang on after the end.
      (%trace3 ";           remove-lines "lines"\n")
      (cond ((null? lines) '())
	    (else
	     (remove-line (car lines))
	     (remove-lines (next-lines (cdr lines))))))

    (define (add-line start num y old)
      (%trace3 ";           add-line "start" "num" "y" "old"\n")
      (let ((new (make-line-ink)))
	(set-line-ink-start! new (mark-permanent-copy start))
	(set-line-ink-end! new (mark-permanent-copy start))
	(set-line-ink-number! new num)
	(fix-drawing-add-ink! drawing new (and (pair? old) (car old)))
	(redraw-line! new 0 y (layout))	;Needs the ink on its drawing.
	(union-ink! new)
	new))

    (define (steal-line! line start num y)
      (%trace3 ";           steal-line! "line" "start" "num" "y"\n")
      (move-mark-to! (line-ink-start line) start)
      (set-line-ink-number! line num)
      (redraw-line! line 0 y (layout))
      (union-ink! line))

    (define (move-line! line start num y)
      (let* ((extent (fix-ink-extent line))
	     (old-num (line-ink-number line))
	     (old-y (fix-rect-y extent)))
	(if (not (fix:= old-y y))
	    (set-text-ink-position! line 0 y))
	(if (not (fix:= old-num num))
	    (set-line-ink-number! line num))
	(if (not (mark= start (line-ink-start line)))
	    (warn "mismatched line-ink start:" start line))
	(union-ink! line)
	(next-y-extent extent)))

    (define (remove-line line)
      (mark-temporary! (line-ink-start line))
      (mark-temporary! (line-ink-end line))
      (without-interruption
       (lambda ()
	 (clear-cached-pango-layout line)
	 (fix-ink-remove! line))))

    (define (no-display-changes?)
      ;; If the drawing already agrees with the buffer and its current
      ;; clipping, return #t.
      (let ((old-tick (buffer-drawing-modified-tick drawing)))
	(and (fix:= old-tick (group-modified-tick group)) ;already redrawn
	     (let ((old-start (buffer-drawing-display-start drawing))
		   (old-end (buffer-drawing-display-end drawing)))
	       (and (mark= old-start display-start)
		    (mark= old-end display-end))))))

    (define (layout)
      (or pango-layout
	  (let ((new (gtk-widget-create-pango-layout
		      (gtk-screen-toplevel screen))))
	    (%trace3 ";             created "new" to lay up new text\n")
	    (pango-layout-set-font-description new (gtk-screen-font screen))
	    (set! pango-layout new)
	    new)))

    (main)))

(define %trace-redraw? #f)

(define (%trace-simplify . args)
  (map (lambda (obj)
	 (cond ((mark? obj) (mark-index obj))
	       ((and (pair? obj) (line-ink? (car obj)))
		(list (car obj) '...))
	       (else obj)))
       args))

(define (redraw-line! line x y pango-layout)
  ;; Updates line by (re)parsing its buffer.  (Re)Images and
  ;; (re)lays-out the line to get its dimensions.  (Re)sizes line and
  ;; (re)positions it at (X, Y).  A separate pango-layout is (re)used
  ;; during this process, and any cached layout is cleared.

  (define-syntax %trace3
    (syntax-rules ()
      ((_ args ...) (if %trace-redraw?
			(apply outf-error (%trace-simplify args ...))))))

  (%trace3 ";\t      redraw-line! "line" from "(line-ink-start line)
	   " ("x","y") with "pango-layout"\n")
  (without-interruption
   (lambda ()
     (%layout-line! line pango-layout)))
  (pango-layout-get-pixel-extents
   pango-layout
   (lambda (width height)
     (without-interruption
      (lambda ()
	(clear-cached-pango-layout line)
	(%trace3 ";\t        erasing "(fix-ink-extent line)"\n")
	(drawing-damage line)
	(let ((extent (fix-ink-extent line)))
	  (set-fix-rect-size! extent width height)
	  (set-fix-rect-position! extent x y))
	(%trace3 ";\t        drawing "(fix-ink-extent line)"\n")
	(drawing-damage line))))))

(define image-buffer-size (* 50 1024))
(define image-buffer (string-allocate image-buffer-size))

(define (%layout-line! line pango-layout)
  (let* ((drawing (fix-ink-drawing line))
	 (buffer (buffer-drawing-buffer drawing))
	 (group (buffer-group buffer))
	 (max-image-size (fix:-1+ image-buffer-size)))
    ;; Image the whole paragraph into a max-sized image-buffer.
    (set-string-length! image-buffer image-buffer-size)
    (gtk-group-line-image!
     group (line-ink-start-index line) (group-display-end-index group)
     image-buffer 0 max-image-size
     (buffer-drawing-tab-width drawing)
     0 ;; column-offset
     (buffer-drawing-char-image-strings drawing)
     (lambda (text-index image-index)
       (if (fix:= image-index max-image-size)
	   (warn ";%layout-line!: long paragraph"))
       (set-mark-index! (line-ink-end line) text-index)

       ;; Run Pango on buffer.
       (if (fix:> image-index 0)
	   (begin
	     (set-string-length! image-buffer image-index)
	     (pango-layout-set-markup pango-layout image-buffer))
	   (pango-layout-set-text pango-layout ""))))))

(define (final-newline? group)
  (let ((index (group-display-end-index group)))
    (and (not (group-start-index? group index))
	 (char=? #\newline (group-left-char group index)))))

(declare (integrate-operator unchanged?))
(define (unchanged? line)
  (let* ((drawing (fix-ink-drawing line))
	 (buffer (buffer-drawing-buffer drawing)))
    (and buffer
	 (let ((group (buffer-group buffer)))
	   (%unchanged? line
			(group-start-changes-index group)
			(group-end-changes-index group))))))

(declare (integrate-operator %unchanged?))
(define (%unchanged? line change-start change-end)
  (or
   ;; Common trivial case: no change = unchanged.
   (not change-start)

   ;; First case: the change region ends before line starts.
   ;;
   ;; Line and change region may not touch.  The change region may
   ;; have removed the newline before line, or inserted new text
   ;; after the newline, changing line's start.
   (let ((line-start (line-ink-start-index line)))
     (fix:< change-end line-start))

   ;; Second case: the change region starts after line ends.
   ;;
   ;; Line must end with a newline, else a change region touching
   ;; the end is adding to the line.  Rather than test for this,
   ;; consider touching lines as NOT unchanged.
   (let ((line-end (line-ink-end-index line)))
     (fix:< line-end change-start))))

(define (update-cursor widget)
  (%trace ";\t  update-cursor "widget"\n")
  (let ((window (text-widget-buffer-frame widget))
	(cursor (text-widget-cursor-ink widget)))
    (%trace ";\t    cursor: "cursor"\n")
    (redraw-cursor widget (window-point window))

    ;; Get cursor appearance right per current mode.  An active
    ;; typein window looks selected, else invisible.  An active buffer
    ;; looks selected, else visible.
    (let ((selected (screen-cursor-window (window-screen window))))
      (cond ((eq? window selected)
	     (visible! cursor #t))
	    ((and (text-widget? widget)
		  (not (text-widget-modeline widget)))
	     (visible! cursor #f))
	    (else ;; text widget
	     (visible! cursor #t))))))

(define (redraw-cursor widget point)
  (%trace ";\t      redraw-cursor at "point" in "widget"\n")
  (let* ((window (text-widget-buffer-frame widget))
	 (screen (window-screen window))
	 (cursor (text-widget-cursor-ink widget))
	 (line (find-line-at point widget))
	 (group (mark-group point)))
    (%trace ";\t\tfound line: "line"\n")

    (define-integrable (main)
      (cond
       ((not cursor)
	(%trace ";\t\tno widget for "window"\n")
	#t)

       ;; When beyond a final newline, position cursor where next line
       ;; would start.
       ((and line
	     (mark= point (group-display-end group))
	     (final-newline? group))
	(let* ((extent (fix-ink-extent line))
	       (line-spacing (gtk-screen-line-spacing screen))
	       (y (fix:+ (fix-rect-max-y extent) line-spacing)))
	  (%trace ";\t\tredraw-cursor beyond final newline, at 0,"y"\n")
	  (set-half-box! 0 y)))

       ;; Else at end (or inside) found line.
       (line
	(let* ((extent (fix-ink-extent line))
	       (layout (text-ink-pango-layout line))
	       (column (image-column point line)))
	  (pango-layout-index-to-pos
	   layout column
	   (lambda (xG yG widthG heightG)
	     (let ((log-x (fix:+ xG (fix-rect-x extent)))
		   (log-y (fix:+ yG (fix-rect-y extent))))
	       (%trace ";\t\tredraw-cursor: index-to-pos: "column
		       " => "log-x","log-y" "widthG"x"heightG" - "layout"\n")
	       (set-box! log-x log-y widthG heightG))))))

       ;; Else... a half-char box for the empty buffer.
       (else
	(%trace ";\t\tno line found: half box at 0,0\n")
	(set-half-box! 0 0))))

    (define (set-half-box! x y)
      (let ((half-width (quotient (gtk-screen-char-width screen) 2))
	    (line-height (gtk-screen-line-height screen)))
	(set-rectangle-ink! cursor x y half-width line-height))
      #t)

    (define (set-box! x y width height)
      (if (fix:< width 5)
	  (set-rectangle-ink! cursor x y 5 height)
	  (set-rectangle-ink! cursor x y width height))
      #t)

    (main)))

(define (image-column point line)
  ;; Returns the index of the character at point within line's image.
  (let* ((drawing (fix-ink-drawing line))
	 (buffer (buffer-drawing-buffer drawing))
	 (group (buffer-group buffer)))
    (group-columns group
		   (mark-index (line-ink-start line))
		   (mark-index point)
		   0 ;; start column
		   (buffer-drawing-tab-width drawing)
		   (buffer-drawing-char-image-strings drawing))))

;;; Buffer Drawings and Buffer Lines

(define-class (<buffer-drawing>
	       (constructor make-buffer-drawing
			    (buffer tab-width char-image-strings)
			    no-init))
    (<fix-drawing>)

  ;; The buffer being drawn, and the "visual" parameters affecting its
  ;; rendition.
  (buffer define accessor)
  (tab-width define accessor)
  (char-image-strings define accessor)

  ;; If the drawing has not been kept up-to-date with the buffer, set
  ;; this flag to #f.  The next redraw will ignore the buffer's change
  ;; region and redraw the entire buffer (and set this back to #t).
  (valid? define standard initial-value #f)

  ;; The buffer's modified-tick, and copies of the buffer's
  ;; display-start/end at the time of the last successful redraw.
  (modified-tick define standard initial-value #f)
  (display-start define standard initial-value #f)
  (display-end define standard initial-value #f)

  ;; These are the particulars of the set of PangoLayouts in use.
  ;; Each element is a "cache" containing: (<line-ink>|#f
  ;; . <pango-layout>).  Thus each layout is either idle, or in use --
  ;; in a line-ink's cached-pango-layout slot.
  (pango-layout-caches define standard initial-value '()))

;; The pango-layout-cache abstraction:
(define-integrable make-cache cons)
(define-integrable cache-line car)
(define-integrable cache-layout cdr)
(define-integrable set-cache-line! set-car!)
(define (find-cache line drawing)
  (or
   (assq line (buffer-drawing-pango-layout-caches drawing))
   (error "missing from pango-layout cache" line drawing)))

(define-class (<line-ink> (constructor ()))
    (<text-ink>)

  (start define standard initial-value #f)
  (end define standard initial-value #f)
  (number define standard initial-value #f)
  (cached-pango-layout define standard initial-value #f))

(define (line-ink-start-index line)
  (let ((mark (line-ink-start line)))
    (and mark (mark-index mark))))

(define (line-ink-end-index line)
  (let ((mark (line-ink-end line)))
    (and mark (mark-index mark))))

(define-method write-instance ((line <line-ink>) port)
  (write-instance-helper
   "line-ink" line port
   (lambda ()
     (write-char #\space port)
     (write-char #\# port)
     (write (line-ink-number line) port)
     (write-char #\space port)
     (write (line-ink-start-index line) port)
     (write-char #\- port)
     (write (line-ink-end-index line) port))))

(define-method initialize-instance ((ink <line-ink>))
  (call-next-method ink)
  (set-text-ink-color! ink "black"))

(define-method text-ink-pango-layout ((ink <line-ink>))
  ;; A cached pango layout is presumed to be all laid out.  A cache
  ;; miss means a PangoLayout must be re-laid-up from the buffer text.

  (define (salvage-pango-layout line)
    ;; Look for a cached PangoLayout to re-use.  Returns abandoned
    ;; layouts (whose line is #f), and layouts for lines that are
    ;; off-screen in all of the drawing's widgets.
    (let* ((drawing (fix-ink-drawing line))
	   (widgets (fix-drawing-widgets drawing)))
      (let loop ((caches (buffer-drawing-pango-layout-caches drawing)))
	(if (null? caches)
	    #f
	    (let* ((cache (car caches))
		   (old (cache-line cache)))
	      (if (or (eq? old #f)
		      (every (let ((old-extent (fix-ink-extent old)))
			       (lambda (widget)
				 (or (gtk-widget-destroyed? widget)
				     (not (fix-rect-intersect?
					   old-extent
					   (fix-layout-view widget))))))
			     widgets))
		  (let ((layout (cache-layout cache)))
		    (if old (set-line-ink-cached-pango-layout! old #f))
		    (set-cache-line! cache line)
		    (set-line-ink-cached-pango-layout! line layout)
		    layout)
		  (loop (cdr caches))))))))

  (define (cache-pango-layout line)
    (let* ((drawing (fix-ink-drawing line))
	   (widget (car (fix-drawing-widgets drawing)))
	   (layout (gtk-widget-create-pango-layout widget))
	   (new (make-cache line layout)))
      (pango-layout-set-font-description layout (gtk-screen-font
						 (edwin-widget-screen widget)))
      (set-buffer-drawing-pango-layout-caches!
       drawing (cons new (buffer-drawing-pango-layout-caches drawing)))
      (set-line-ink-cached-pango-layout! line layout)
      layout))

  ;; Do not (call-next-method ink).  There is no <text-ink> method.
  (or (line-ink-cached-pango-layout ink)
      (let ((layout (or (salvage-pango-layout ink)
			(cache-pango-layout ink))))
	(%layout-line! ink layout)
	layout)))

(define (clear-cached-pango-layout line)
  (let ((layout (line-ink-cached-pango-layout line)))
    (if layout
	(let* ((drawing (fix-ink-drawing line))
	       (cache (find-cache line drawing)))
	  (set-cache-line! cache #f)
	  (set-line-ink-cached-pango-layout! line #f)))))

(define-class (<cursor-ink> (constructor ()))
    (<rectangle-ink>)

  ;; #t if the cursor should be drawn.
  (visible? define standard initial-value #t)

  ;; A list of one <fix-layout>.  Used to blink this ink "on"
  ;; (restore its ink-widgets list) withOUT consing.
  (widget-list define standard))

(define-method initialize-instance ((cursor <cursor-ink>))
  (call-next-method cursor)
  (set-rectangle-ink-width! cursor 2)
  (set-rectangle-ink-width! cursor 2)
  (set-rectangle-ink-color! cursor "green")
  (set-rectangle-ink-fill-color! cursor (make-rgba-color 0. 1. 0. .5)))

(define (make-rgba-color red green blue alpha)
  (let ((color (flo:vector-cons 4)))
    (flo:vector-set! color 0 red)
    (flo:vector-set! color 1 green)
    (flo:vector-set! color 2 blue)
    (flo:vector-set! color 3 alpha)
    color))

(define (guarantee-cursor-ink object)
  (if (cursor-ink? object) object
      (error:wrong-type-argument object "<cursor-ink>" 'guarantee-cursor-ink)))

(define (cursor-ink-widget cursor)
  (car (cursor-ink-widget-list cursor)))

(define (visible! cursor visible?)
  ;; Atomically sets cursor-ink-visible? and fix-ink-widgets.
  (without-interruption
   (lambda ()
     (if visible?
	 (if (not (cursor-ink-visible? cursor))
	     (begin
	       (set-fix-ink-widgets! cursor (cursor-ink-widget-list cursor))
	       (set-cursor-ink-visible?! cursor #t)))
	 (if (cursor-ink-visible? cursor)
	     (begin
	       (set-cursor-ink-visible?! cursor #f)
	       (set-fix-ink-widgets! cursor '())))))))

(define (blink! screen cursor)
  ;; Atomically sets cursor up to blink.  Cursor may be #f, in which
  ;; case blinking will pause.
  (without-interruption
   (lambda ()
     (let ((old (gtk-screen-blinking screen)))
       (if cursor
	   (begin
	     (if (not (eq? cursor old))
		 (set-gtk-screen-blinking! screen cursor))
	     (if (not old)
		 (signal-thread-event (gtk-screen-blinker screen)
				      (lambda () #f))))
	   (if old (set-gtk-screen-blinking! screen #f)))))))

;;; Buffer Status
;;
;; The (re)layout process starts at the top of a changed buffer region
;; and works its way to the bottom, scanning for line separators,
;; "imaging" the content (e.g. replacing #\null with "^@"), feeding
;; the translation (with style info!) to Pango for layup, and stacking
;; the laid-up lines.  Reading a large file may produce a change
;; region containing hundreds of thousands of lines, taking a
;; non-interactive amount of time to layout for display.  In spite of
;; this, the user may want to type ahead, e.g. go to the end of the
;; buffer and start typing in a new line.
;;
;; To keep redisplay interactive in such a case, a thread might be
;; spawned to do the layout.  The editor thread can then continue with
;; event (keypress) processing.  The new thread works on the buffer
;; (re)drawing, and shows its progress by animating a progress/status
;; indicator in the drawing.  The indicator might report the number of
;; bytes remaining to be (re)parsed, with newly re-parsed lines
;; appearing above it.  When the point is at buffer indices that are
;; not (yet) laid out, the cursor appears after the progress
;; indicator.  Any typeahead will be displayed... eventually.

(define-class (<buffer-status> (constructor add-buffer-status (drawing) 1))
    (<rectangle-ink>)
  (text-ink define standard))

(define (with-interruption thunk)
  (%trace "; with-interruption "thunk"\n")
  (assert-without-interruption 'with-interruption)
  (unblock-thread-events)
  (let ((v (thunk)))
    (%trace "; with-interruption "thunk" => "v"\n")
    (block-thread-events)
    v))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ args ...) (if %trace? (outf-error args ...)))))

(define %trace2? #f)

(define-syntax %trace2
  (syntax-rules ()
    ((_ args ...) (if %trace2? (outf-error args ...)))))

(define %trace-buttons? #f)

(define-syntax %trace-buttons
  (syntax-rules ()
    ((_ args ...) (if %trace-buttons? (%%trace-buttons args ...)))))

(define (%%trace-buttons . msg)
  (apply outf-error `("; Button ",@msg"\n")))

(initialize-package!)